# Copyright 2023-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
These commands are used by pre-commit.
"""
# pylint: disable=resource-leakage,broad-except,3rd-party-module-not-gated
from __future__ import annotations

import shutil
from typing import NoReturn

from ptscripts import command_group
from ptscripts import Context


# Define the command group
cgroup = command_group(name="pre-commit", help="Pre-Commit Related Commands", description=__doc__)


@cgroup.command(  # type: ignore[misc]
    name="actionlint",
    arguments={
        "files": {
            "help": "Files to run actionlint against",
            "nargs": "*",
        },
        "no_color": {
            "help": "Disable colors in output",
        },
    },
)
def actionlint(ctx: Context, files: list[str], no_color: bool = False) -> NoReturn:
    """
    Run `actionlint`.
    """
    actionlint = shutil.which("actionlint")
    if not actionlint:
        ctx.warn("Could not find the 'actionlint' binary")
        ctx.exit(0)
    cmdline = [actionlint]
    if no_color is False:
        cmdline.append("-color")
    shellcheck = shutil.which("shellcheck")
    if shellcheck:
        cmdline.append(f"-shellcheck={shellcheck}")
    pyflakes = shutil.which("pyflakes")
    if pyflakes:
        cmdline.append(f"-pyflakes={pyflakes}")
    ret = ctx.run(*cmdline, *files, check=False)
    ctx.exit(ret.returncode)



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import functools
import logging
import pprint
import re
import sys
import time
from typing import List

import attr
import psutil
import pytest
from pytestskipmarkers.utils import platform

from pytestshellutils.exceptions import FactoryNotRunning
from pytestshellutils.exceptions import FactoryNotStarted
from pytestshellutils.shell import Daemon
from pytestshellutils.utils.processes import _get_cmdline
from tests.conftest import Tempfiles

try:
    from pytest import FixtureRequest
    from pytest import LogCaptureFixture
except ImportError:
    from _pytest.fixtures import FixtureRequest
    from _pytest.logging import LogCaptureFixture

PROCESS_START_TIMEOUT = 2

log = logging.getLogger(__name__)


def kill_children(procs: List[psutil.Process]) -> None:  # pragma: no cover
    _, alive = psutil.wait_procs(procs, timeout=3)
    for p in alive:
        p.kill()


def test_daemon_process_termination(request: FixtureRequest, tempfiles: Tempfiles) -> None:
    primary_childrend_count = 5
    secondary_children_count = 3
    script = tempfiles.makepyfile(
        """
        #!{shebang}
        # coding=utf-8

        import time
        import multiprocessing

        def spin():
            while True:
                try:
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    break

        def spin_children():
            procs = []
            for idx in range({secondary_children_count}):
                proc = multiprocessing.Process(target=spin)
                proc.daemon = True
                proc.start()
                procs.append(proc)

            while True:
                try:
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    break


        def main():
            procs = []

            for idx in range({primary_childrend_count}):
                proc = multiprocessing.Process(target=spin_children)
                procs.append(proc)
                proc.start()

            while True:
                try:
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    break

            # We're not terminating child processes on purpose. Our code should handle it.

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """.format(
            shebang=sys.executable,
            primary_childrend_count=primary_childrend_count,
            secondary_children_count=secondary_children_count,
        ),
        executable=True,
    )
    if not platform.is_windows():
        daemon = Daemon(start_timeout=1, script_name=script)
    else:  # pragma: is-windows
        # Windows don't know how to handle python scripts directly
        daemon = Daemon(start_timeout=1, script_name=sys.executable, base_script_args=[script])
    daemon.start()
    daemon_pid = daemon.pid
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)
    # Allow the script to start
    time.sleep(PROCESS_START_TIMEOUT)
    assert psutil.pid_exists(daemon_pid)
    proc = psutil.Process(daemon_pid)
    children = proc.children(recursive=True)
    request.addfinalizer(functools.partial(kill_children, children))
    child_count = len(children)
    expected_count = primary_childrend_count + (primary_childrend_count * secondary_children_count)
    if platform.is_windows() and sys.version_info >= (3, 7):  # pragma: is-windows-ge-py37
        # After Python 3.7 there's an extra spawning process
        expected_count += 1
    if platform.is_darwin() and sys.version_info >= (3, 8):  # pragma: is-darwin-ge-py38
        # macOS defaults to spawning new processed after Python 3.8
        # Account for the forking process
        expected_count += 1
    assert child_count == expected_count, "{}!={}\n{}".format(
        child_count,
        expected_count,
        pprint.pformat([_get_cmdline(child) or child for child in children]),
    )
    daemon.terminate()
    assert psutil.pid_exists(daemon_pid) is False
    for child in list(children):  # pragma: no cover
        if psutil.pid_exists(child.pid):
            continue
        children.remove(child)
    assert not children, "len(children)=={} != 0\n{}".format(
        len(children), pprint.pformat([_get_cmdline(child) or child for child in children])
    )


@pytest.mark.skip("Will debug later")
def test_daemon_process_termination_parent_killed(
    request: FixtureRequest, tempfiles: Tempfiles
) -> None:

    primary_childrend_count = 5
    secondary_children_count = 3
    script = tempfiles.makepyfile(
        """
        #!{shebang}
        # coding=utf-8

        import time
        import multiprocessing

        def spin():
            while True:
                try:
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    break

        def spin_children():
            procs = []
            for idx in range({secondary_children_count}):
                proc = multiprocessing.Process(target=spin)
                proc.daemon = True
                proc.start()
                procs.append(proc)

            while True:
                try:
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    break

        def main():
            procs = []

            for idx in range({primary_childrend_count}):
                proc = multiprocessing.Process(target=spin_children)
                procs.append(proc)
                proc.start()

            while True:
                try:
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    break

            # We're not terminating child processes on purpose. Our code should handle it.

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """.format(
            shebang=sys.executable,
            primary_childrend_count=primary_childrend_count,
            secondary_children_count=secondary_children_count,
        ),
        executable=True,
    )
    if not platform.is_windows():
        daemon = Daemon(start_timeout=1, script_name=script)
    else:  # pragma: is-windows
        # Windows don't know how to handle python scripts directly
        daemon = Daemon(start_timeout=1, script_name=sys.executable, base_script_args=[script])
    daemon.start()
    daemon_pid = daemon.pid
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)
    # Allow the script to start
    time.sleep(PROCESS_START_TIMEOUT)
    assert psutil.pid_exists(daemon_pid)
    proc = psutil.Process(daemon_pid)
    children = proc.children(recursive=True)
    request.addfinalizer(functools.partial(kill_children, children))
    assert len(children) == primary_childrend_count + (
        primary_childrend_count * secondary_children_count
    )
    # Pretend the parent process died.
    proc.kill()
    time.sleep(0.5)
    # We should should still be able to terminate all child processes
    daemon.terminate()
    assert psutil.pid_exists(daemon_pid) is False
    psutil.wait_procs(children, timeout=3)
    for child in list(children):
        if psutil.pid_exists(child.pid):
            continue
        children.remove(child)
    assert not children, "len(children)=={} != 0\n{}".format(
        len(children), pprint.pformat(children)
    )


@pytest.mark.parametrize("start_timeout", [0.1, 0.3])
def test_started_context_manager(
    request: FixtureRequest, tempfiles: Tempfiles, start_timeout: float
) -> None:
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8

        import sys
        import time
        import multiprocessing

        def main():
            time.sleep(3)
            sys.stdout.write("Done!\n")
            sys.stdout.flush()
            sys.exit(0)

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """,
        executable=True,
    )
    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[script],
        start_timeout=2,
        max_start_attempts=1,
        check_ports=[12345],
    )
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)
    with pytest.raises(FactoryNotStarted) as exc:
        daemon.start(start_timeout=start_timeout)
    match = re.search(r"which took (?P<seconds>.*) seconds", str(exc.value))
    assert match
    # XXX: Revisit logic
    # seconds = float(match.group("seconds"))
    # Must take at least start_timeout to start
    # assert seconds > start_timeout
    # Should not take more than start_timeout + 0.3 to start and fail
    # assert seconds < start_timeout + 0.3

    # And using a context manager?
    with pytest.raises(FactoryNotStarted) as exc:
        started = None
        with daemon.started(start_timeout=start_timeout):
            # We should not even be able to set the following variable
            started = False  # pragma: no cover
    assert started is None
    match = re.search(r"which took (?P<seconds>.*) seconds", str(exc.value))
    assert match
    # XXX: Revisit logic
    # seconds = float(match.group("seconds"))
    # Must take at least start_timeout to start
    # assert seconds > start_timeout
    # Should not take more than start_timeout + 0.3 to start and fail
    # assert seconds < start_timeout + 0.3


@pytest.fixture
def factory_stopped_script(tempfiles: Tempfiles) -> str:
    return tempfiles.makepyfile(
        r"""
        # coding=utf-8

        import os
        import sys
        import time
        import socket
        import multiprocessing

        def main():

            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind(('127.0.0.1', 12345))
            sock.listen(5)
            try:
                while True:
                    connection, address = sock.accept()
                    connection.close()
            except (KeyboardInterrupt, SystemExit):
                pass
            finally:
                sock.close()
            sys.exit(0)

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """,
        executable=True,
    )


def test_stopped_context_manager_raises_FactoryNotRunning(
    request: FixtureRequest, factory_stopped_script: str
) -> None:
    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[factory_stopped_script],
        start_timeout=3,
        max_start_attempts=1,
        check_ports=[12345],
    )
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)

    with pytest.raises(FactoryNotRunning):
        with daemon.stopped():
            pass  # pragma: no cover


def test_stopped_context_manager(request: FixtureRequest, factory_stopped_script: str) -> None:
    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[factory_stopped_script],
        start_timeout=3,
        max_start_attempts=1,
        check_ports=[12345],
    )
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)

    with daemon.started():
        assert daemon.is_running()
        with daemon.stopped():
            assert daemon.is_running() is False
        assert daemon.is_running()


@attr.s
class DaemonCallbackCounter:
    before_start_callback_counter = attr.ib(default=0)  # type: int
    after_start_callback_counter = attr.ib(default=0)  # type: int
    before_terminate_callback_counter = attr.ib(default=0)  # type: int
    after_terminate_callback_counter = attr.ib(default=0)  # type: int

    def before_start_callback(self) -> None:
        self.before_start_callback_counter += 1

    def after_start_callback(self) -> None:
        self.after_start_callback_counter += 1

    def before_terminate_callback(self) -> None:
        self.before_terminate_callback_counter += 1

    def after_terminate_callback(self) -> None:
        self.after_terminate_callback_counter += 1


@attr.s
class DaemonContextCallbackCounter:
    daemon = attr.ib()  # type: Daemon
    before_start_callback_counter = attr.ib(default=0)  # type: int
    after_start_callback_counter = attr.ib(default=0)  # type: int
    before_stop_callback_counter = attr.ib(default=0)  # type: int
    after_stop_callback_counter = attr.ib(default=0)  # type: int

    def before_start_callback(self, daemon: Daemon) -> None:
        assert daemon is self.daemon
        self.before_start_callback_counter += 1

    def after_start_callback(self, daemon: Daemon) -> None:
        assert daemon is self.daemon
        self.after_start_callback_counter += 1

    def before_stop_callback(self, daemon: Daemon) -> None:
        assert daemon is self.daemon
        self.before_stop_callback_counter += 1

    def after_stop_callback(self, daemon: Daemon) -> None:
        assert daemon is self.daemon
        self.after_stop_callback_counter += 1


def test_daemon_callbacks(request: FixtureRequest, factory_stopped_script: str) -> None:

    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[factory_stopped_script],
        start_timeout=3,
        max_start_attempts=1,
        check_ports=[12345],
    )
    callbacks = DaemonCallbackCounter()
    daemon.before_start(callbacks.before_start_callback)
    daemon.after_start(callbacks.after_start_callback)
    daemon.before_terminate(callbacks.before_terminate_callback)
    daemon.after_terminate(callbacks.after_terminate_callback)

    stopped_callbacks = DaemonContextCallbackCounter(daemon)

    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)

    daemon_started_once = False

    with daemon.started():
        daemon_started_once = daemon.is_running()
        assert daemon_started_once is True

        # Assert against the non context manager callbacks
        assert callbacks.before_start_callback_counter == 1
        assert callbacks.after_start_callback_counter == 1
        assert callbacks.before_terminate_callback_counter == 0
        assert callbacks.after_terminate_callback_counter == 0

        # Assert against the context manager callbacks
        assert stopped_callbacks.before_stop_callback_counter == 0
        assert stopped_callbacks.after_stop_callback_counter == 0
        assert stopped_callbacks.before_start_callback_counter == 0
        assert stopped_callbacks.after_start_callback_counter == 0

        with daemon.stopped(
            before_stop_callback=stopped_callbacks.before_stop_callback,
            after_stop_callback=stopped_callbacks.after_stop_callback,
            before_start_callback=stopped_callbacks.before_start_callback,
            after_start_callback=stopped_callbacks.after_start_callback,
        ):
            assert daemon.is_running() is False

            # Assert against the context manager callbacks
            assert stopped_callbacks.before_stop_callback_counter == 1
            assert stopped_callbacks.after_stop_callback_counter == 1
            assert stopped_callbacks.before_start_callback_counter == 0
            assert stopped_callbacks.after_start_callback_counter == 0

            # Assert against the non context manager callbacks
            assert callbacks.before_start_callback_counter == 1
            assert callbacks.after_start_callback_counter == 1
            assert callbacks.before_terminate_callback_counter == 1
            assert callbacks.after_terminate_callback_counter == 1

        assert daemon.is_running()

        # Assert against the context manager callbacks
        assert stopped_callbacks.before_stop_callback_counter == 1
        assert stopped_callbacks.after_stop_callback_counter == 1
        assert stopped_callbacks.before_start_callback_counter == 1
        assert stopped_callbacks.after_start_callback_counter == 1

        # Assert against the non context manager callbacks
        assert callbacks.before_start_callback_counter == 2
        assert callbacks.after_start_callback_counter == 2
        assert callbacks.before_terminate_callback_counter == 1
        assert callbacks.after_terminate_callback_counter == 1

        # Let's got through stopped again, the stopped_callbacks should not be called again
        # because they are not passed into .stopped()
        with daemon.stopped():
            assert daemon.is_running() is False
        assert daemon.is_running()

        # Assert against the context manager callbacks
        assert stopped_callbacks.before_stop_callback_counter == 1
        assert stopped_callbacks.after_stop_callback_counter == 1
        assert stopped_callbacks.before_start_callback_counter == 1
        assert stopped_callbacks.after_start_callback_counter == 1

    assert daemon_started_once is True

    # Assert against the non context manager callbacks
    assert callbacks.before_start_callback_counter == 3
    assert callbacks.after_start_callback_counter == 3
    assert callbacks.before_terminate_callback_counter == 3
    assert callbacks.after_terminate_callback_counter == 3


@attr.s
class DaemonStartCheckCounter:
    custom_start_check_1_callback_counter = attr.ib(default=0)  # type: int
    custom_start_check_2_callback_counter = attr.ib(default=0)  # type: int
    custom_start_check_3_callback_counter = attr.ib(default=0)  # type: int

    def custom_start_check_1_callback(self, timeout_at: float) -> bool:
        self.custom_start_check_1_callback_counter += 1
        if self.custom_start_check_1_callback_counter > 2:
            return True
        return False

    def custom_start_check_2_callback(self, timeout_at: float) -> bool:
        self.custom_start_check_2_callback_counter += 1
        if self.custom_start_check_2_callback_counter > 2:
            return True
        raise Exception("Foo!")

    def custom_start_check_3_callback(self, timeout_at: float) -> bool:
        self.custom_start_check_3_callback_counter += 1
        time.sleep(1)
        return False


def test_daemon_start_check_callbacks(request: FixtureRequest, factory_stopped_script: str) -> None:

    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[factory_stopped_script],
        start_timeout=3,
        max_start_attempts=1,
        check_ports=[12345],
    )
    callbacks = DaemonStartCheckCounter()
    daemon.start_check(callbacks.custom_start_check_1_callback)
    daemon.start_check(callbacks.custom_start_check_2_callback)

    daemon_start_check_callbacks = daemon.get_start_check_callbacks()

    with daemon.started():
        # Both start callbacks should have run 3 times by now, at which
        # time, they would have returned True
        pass

    assert callbacks.custom_start_check_1_callback_counter == 3
    assert callbacks.custom_start_check_2_callback_counter == 3

    # Assert that the list of callbacks is the same before running the start checks
    assert daemon.get_start_check_callbacks() == daemon_start_check_callbacks


def test_daemon_no_start_check_callbacks(request: FixtureRequest, tempfiles: Tempfiles) -> None:
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8

        import sys
        import time
        import multiprocessing

        def main():
            time.sleep(3)
            sys.stdout.write("Done!\n")
            sys.stdout.flush()
            sys.exit(0)

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """,
        executable=True,
    )
    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[script],
        start_timeout=2,
        max_start_attempts=1,
    )
    # Remove the check ports callback
    daemon._start_checks_callbacks.clear()
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)
    with daemon.started():
        # Daemon started without running any start checks
        pass
    assert not daemon.get_start_check_callbacks()


def test_daemon_start_check_callbacks_factory_not_running(
    request: FixtureRequest, tempfiles: Tempfiles
) -> None:
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8

        import sys
        import time
        import multiprocessing

        def main():
            time.sleep(2)
            sys.stdout.write("Done!\n")
            sys.stdout.flush()
            sys.exit(0)

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """,
        executable=True,
    )

    callbacks = DaemonStartCheckCounter()

    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[script],
        start_timeout=2,
        max_start_attempts=1,
    )
    # Make sure the daemon is terminated no matter what
    request.addfinalizer(daemon.terminate)

    daemon.start_check(callbacks.custom_start_check_3_callback)
    with pytest.raises(FactoryNotStarted):
        daemon.start()
    # Make sure the callback was called at least once
    assert callbacks.custom_start_check_3_callback_counter > 1


def test_context_manager_returns_class_instance(tempfiles: Tempfiles) -> None:
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8

        import sys
        import time
        import multiprocessing

        def main():
            while True:
                try:
                    time.sleep(0.1)
                except KeyboardInterrupt:
                    break
            sys.stdout.write("Done!\n")
            sys.stdout.flush()
            sys.exit(0)

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """,
        executable=True,
    )
    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[script],
        start_timeout=1,
        max_start_attempts=1,
    )

    # Without starting the factory
    started = d = None
    with pytest.raises(RuntimeError):
        with daemon as d:
            # We should not even be able to set the following variable
            started = d.is_running()  # pragma: no cover
    assert d is None
    assert started is None

    # After starting the factory
    started = False
    daemon.start()
    with daemon as d:
        # We should not even be able to set the following variable
        started = d.is_running()
    assert d.is_running() is False
    assert started is True

    # By starting the factory and passing timeout directly
    started = False
    with daemon.started(start_timeout=1) as d:
        # We should not even be able to set the following variable
        started = d.is_running()
    assert d.is_running() is False
    assert started is True

    # By starting the factory without any keyword arguments
    started = False
    with daemon.started() as d:
        # We should not even be able to set the following variable
        started = d.is_running()
    assert d.is_running() is False
    assert started is True


@pytest.mark.parametrize("max_start_attempts", [1, 2, 3])
def test_exact_max_start_attempts(
    tempfiles: Tempfiles, caplog: LogCaptureFixture, max_start_attempts: int
) -> None:
    """
    This test asserts that we properly report max_start_attempts.
    """
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8

        import sys
        import time
        import multiprocessing

        def main():
            time.sleep(0.125)
            sys.exit(1)

        # Support for windows test runs
        if __name__ == '__main__':
            multiprocessing.freeze_support()
            main()
        """,
        executable=True,
    )
    daemon = Daemon(
        script_name=sys.executable,
        base_script_args=[script],
        start_timeout=0.1,
        max_start_attempts=max_start_attempts,
        check_ports=[12345],
    )
    with caplog.at_level(logging.INFO):
        with pytest.raises(FactoryNotStarted) as exc:
            daemon.start()
        assert "confirm running status after {} attempts".format(max_start_attempts) in str(
            exc.value
        )
    start_attempts = [
        "Attempt: {} of {}".format(n, max_start_attempts) for n in range(1, max_start_attempts + 1)
    ]
    for record in caplog.records:
        if not record.message.startswith("Starting Daemon"):
            continue
        for idx, start_attempt in enumerate(list(start_attempts)):
            if start_attempt in record.message:
                start_attempts.pop(idx)
    assert not start_attempts



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Shelling class implementations.
"""
import atexit
import contextlib
import json
import locale
import logging
import os
import pathlib
import shutil
import subprocess
import sys
from tempfile import SpooledTemporaryFile
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

import attr
import psutil
from pytestskipmarkers.utils import platform

from pytestshellutils.customtypes import Callback
from pytestshellutils.customtypes import EnvironDict
from pytestshellutils.exceptions import CallbackException
from pytestshellutils.exceptions import FactoryNotRunning
from pytestshellutils.exceptions import FactoryNotStarted
from pytestshellutils.exceptions import FactoryTimeout
from pytestshellutils.exceptions import ShellUtilsException
from pytestshellutils.utils import format_callback_to_string
from pytestshellutils.utils import ports
from pytestshellutils.utils import resolved_pathlib_path
from pytestshellutils.utils import time
from pytestshellutils.utils.processes import ProcessResult
from pytestshellutils.utils.processes import terminate_process
from pytestshellutils.utils.processes import terminate_process_list

if TYPE_CHECKING:
    from typing import Type
    from pytestsysstats.plugin import StatsProcesses

log = logging.getLogger(__name__)


@attr.s(slots=True, kw_only=True)
class BaseFactory:
    """
    Base factory class.

    Keyword Arguments:
        cwd:
            The path to the desired working directory
        environ:
            A dictionary of ``key``, ``value`` pairs to add to the environment.
    """

    cwd: pathlib.Path = attr.ib(converter=resolved_pathlib_path)
    environ: EnvironDict = attr.ib(repr=False)

    @cwd.default
    def _default_cwd(self) -> pathlib.Path:
        """
        Return the default cwd to use.
        """
        return pathlib.Path.cwd()

    @environ.default
    def _default_environ(self) -> EnvironDict:
        """
        Return the default ``os.environ`` to use.
        """
        return cast(EnvironDict, os.environ.copy())


@attr.s(slots=True, kw_only=True)
class SubprocessImpl:
    """
    Subprocess interaction implementation.

    Arguments:
        factory:
            The factory instance, either :py:class:`~pytestshellutils.shell.Subprocess` or
            a sub-class of it.
    """

    factory: "Union[Factory, Subprocess, ScriptSubprocess]" = attr.ib()

    _terminal: "Optional[subprocess.Popen[Any]]" = attr.ib(repr=False, init=False, default=None)
    _terminal_stdout: "Optional[SpooledTemporaryFile[bytes]]" = attr.ib(
        repr=False, init=False, default=None
    )
    _terminal_stderr: "Optional[SpooledTemporaryFile[bytes]]" = attr.ib(
        repr=False, init=False, default=None
    )
    _terminal_result: Optional[ProcessResult] = attr.ib(repr=False, init=False, default=None)
    _terminal_timeout: Union[int, float] = attr.ib(repr=False, init=False, default=None)
    _children: List[psutil.Process] = attr.ib(repr=False, init=False, factory=list)

    def cmdline(self, *args: str, **kwargs: Any) -> List[str]:
        """
        Construct a list of arguments to use when starting the subprocess.

        Arguments:
            args:
                Additional arguments to use when starting the subprocess

        By default, this method will just call it's factory's ``cmdline()``
        method, but can be overridden.
        """
        return self.factory.cmdline(*args)

    def init_terminal(
        self,
        cmdline: List[str],
        shell: bool = False,
        env: Optional[EnvironDict] = None,
        cwd: Optional[Union[str, pathlib.Path]] = None,
    ) -> "subprocess.Popen[Any]":
        """
        Instantiate a terminal with the passed command line(``cmdline``) and return it.

        Additionally, it sets a reference to it in ``self._terminal`` and also collects
        an initial listing of child processes which will be used when terminating the
        terminal

        Arguments:
            cmdline:
                List of strings to pass as ``args`` to :py:class:`~subprocess.Popen`

        Keyword Arguments:
            shell:
                Pass the value of ``shell`` to :py:class:`~subprocess.Popen`
            env:
                A dictionary of ``key``, ``value`` pairs to add to the
                :py:attr:`pytestshellutils.shell.Factory.environ`.
            cwd:
                A path for the CWD when running the process.

        Returns:
            A :py:class:`~subprocess.Popen` instance.
        """
        environ = self.factory.environ.copy()
        if env is not None:
            environ.update(env)
        self._terminal_stdout = SpooledTemporaryFile(512000, buffering=0)
        self._terminal_stderr = SpooledTemporaryFile(512000, buffering=0)
        close_fds: bool
        if platform.is_windows():  # pragma: is-windows
            # Windows does not support closing FDs
            close_fds = False
        elif platform.is_freebsd() and sys.version_info < (3, 9):  # pragma: is-bsd-lt-py39
            # Closing FDs in FreeBSD before Py3.9 can be slow
            #   https://bugs.python.org/issue38061
            close_fds = False
        else:
            close_fds = True
        self._terminal = subprocess.Popen(
            cmdline,
            stdout=self._terminal_stdout,
            stderr=self._terminal_stderr,
            shell=shell,  # nosec B602
            cwd=str(cwd or self.factory.cwd),
            universal_newlines=True,
            close_fds=close_fds,
            env=environ,
            bufsize=0,
        )
        # Reset the previous _terminal_result if set
        self._terminal_result = None
        try:
            # Check if the process starts properly
            self._terminal.wait(timeout=0.05)
            # If TimeoutExpired is not raised, it means the process failed to start
        except subprocess.TimeoutExpired:
            # We're good
            # Collect any child processes, though, this early there likely is none
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
                for child in psutil.Process(self._terminal.pid).children(
                    recursive=True
                ):  # pragma: no cover
                    if child not in self._children:
                        self._children.append(child)
            atexit.register(self.terminate)
        return self._terminal

    def is_running(self) -> bool:
        """
        Returns true if the sub-process is alive.

        Returns:
            Returns true if the sub-process is alive
        """
        if not self._terminal:
            return False
        return self._terminal.poll() is None

    def terminate(self) -> ProcessResult:
        """
        Terminate the started subprocess.
        """
        return self._terminate()

    def _terminate(self) -> ProcessResult:
        """
        This method actually terminates the started subprocess.
        """
        if self._terminal is None:
            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_result
            return self._terminal_result
        atexit.unregister(self.terminate)
        log.info("Stopping %s", self.factory)
        # Collect any child processes information before terminating the process
        with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
            for child in psutil.Process(self._terminal.pid).children(recursive=True):
                if child not in self._children:
                    self._children.append(child)

        with self._terminal:
            try:
                if self.factory.slow_stop:
                    self._terminal.terminate()
                else:
                    self._terminal.kill()
                try:
                    # Allow the process to exit by itself in case slow_stop is True
                    self._terminal.wait(10)
                except subprocess.TimeoutExpired:  # pragma: no cover
                    # The process failed to stop, no worries, we'll make sure it exit along with it's
                    # child processes bellow
                    pass
            except ProcessLookupError:
                # The process is already gone
                pass
            # Lets log and kill any child processes left behind, including the main subprocess
            # If it failed to properly stop
            terminate_process(
                pid=self._terminal.pid,
                kill_children=True,
                children=self._children,
                slow_stop=self.factory.slow_stop,
            )
            # Wait for the process to terminate, to avoid zombies.
            self._terminal.wait()
            # poll the terminal so the right returncode is set on the popen object
            self._terminal.poll()
            # This call shouldn't really be necessary
            self._terminal.communicate()

            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_stdout
            self._terminal_stdout.flush()
            self._terminal_stdout.seek(0)
            _read_stdout = self._terminal_stdout.read()
            stdout = self._terminal._translate_newlines(  # type: ignore[attr-defined]
                _read_stdout,
                self.factory.system_encoding,
                sys.stdout.errors,
            )
            self._terminal_stdout.close()

            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_stderr
            self._terminal_stderr.flush()
            self._terminal_stderr.seek(0)
            _read_stderr = self._terminal_stderr.read()
            stderr = self._terminal._translate_newlines(  # type: ignore[attr-defined]
                _read_stderr,
                self.factory.system_encoding,
                sys.stderr.errors,
            )
            self._terminal_stderr.close()
        try:
            self._terminal_result = ProcessResult(
                returncode=self._terminal.returncode,
                stdout=stdout,
                stderr=stderr,
                cmdline=cast(List[str], self._terminal.args),
            )
            log.info("%s %s", self.factory.__class__.__name__, self._terminal_result)
            return self._terminal_result
        finally:
            self._terminal = None
            self._terminal_stdout = None
            self._terminal_stderr = None
            self._children = []

    @property
    def pid(self) -> Optional[int]:
        """
        The pid of the running process. None if not running.
        """
        if not self._terminal:  # pragma: no cover
            return None
        return self._terminal.pid

    def run(
        self,
        *args: str,
        shell: bool = False,
        env: Optional[EnvironDict] = None,
        cwd: Optional[Union[str, pathlib.Path]] = None,
        **kwargs: Any,
    ) -> "subprocess.Popen[Any]":
        """
        Run the given command synchronously.

        Arguments:
            args:
                The command to run.

        Keyword Arguments:
            shell:
                Pass the value of `shell` to
                :py:meth:`pytestshellutils.shell.Factory.init_terminal`
            env:
                A dictionary of ``key``, ``value`` pairs to add to the
                :py:attr:`pytestshellutils.shell.Factory.environ`.
            cwd:
                A path for the CWD when running the process.

        Returns:
            A :py:class:`~subprocess.Popen` instance.
        """
        cmdline = self.cmdline(*args, **kwargs)
        log.info("%s is running %r in CWD: %s ...", self.factory, cmdline, cwd or self.factory.cwd)
        return self.init_terminal(cmdline, shell=shell, env=env, cwd=cwd)


@attr.s(slots=True, kw_only=True)
class Factory(BaseFactory):
    """
    Base shell factory class.

    Keyword Arguments:
        slow_stop:
            Whether to terminate the processes by sending a :py:attr:`SIGTERM` signal or by calling
            :py:meth:`~subprocess.Popen.terminate` on the sub-process.
            When code coverage is enabled, one will want `slow_stop` set to `True` so that coverage data
            can be written down to disk.
        system_encoding:
            The system encoding to use when decoding the subprocess output. Defaults to "utf-8".
        timeout:
            The default maximum amount of seconds that a script should run.
            This value can be overridden when calling :py:meth:`~pytestshellutils.shell.Process.run` through
            the ``_timeout`` keyword argument, and, in that case, the timeout value applied would be that
            of ``_timeout`` instead of ``self.timeout``.
    """

    slow_stop: bool = attr.ib(default=True)
    system_encoding: str = attr.ib(repr=False)
    timeout: Union[int, float] = attr.ib()

    impl: SubprocessImpl = attr.ib(repr=False, init=False)

    # Internal attributes
    _cmdline: List[Any] = attr.ib(repr=False, init=False, default=None)

    @system_encoding.default
    def _default_system_encoding(self) -> str:
        return self._get_default_system_encoding()

    @timeout.default
    def _set_timeout(self) -> Optional[int]:
        return self._get_default_timeout()

    def _get_default_system_encoding(self) -> str:
        encoding: Optional[str] = None

        if not platform.is_windows() and sys.stdin is not None:
            # On Linux we can rely on sys.stdin for the encoding since it
            # most commonly matches the file-system encoding. This however
            # does not apply to windows
            encoding = sys.stdin.encoding

        if not encoding:
            # If the system is properly configured this should return a valid
            # encoding. MS Windows has problems with this and reports the wrong
            # encoding

            try:
                encoding = locale.getencoding()  # type: ignore[attr-defined]
            except AttributeError:
                # Python < 3.11
                encoding = locale.getpreferredencoding(do_setlocale=True)

            if not encoding:
                # This is most likely ascii which is not the best but we were
                # unable to find a better encoding. If this fails, we fall all
                # the way back to ascii
                encoding = sys.getdefaultencoding()
            if not encoding:
                if platform.is_darwin():
                    # Mac OS X uses UTF-8
                    encoding = "utf-8"
                elif platform.is_windows():
                    # Windows uses a configurable encoding; on Windows, Python uses the name “mbcs”
                    # to refer to whatever the currently configured encoding is.
                    encoding = "mbcs"
                else:
                    # On linux default to ascii as a last resort
                    encoding = "ascii"

        if not encoding:
            # If we still didn't detect the encoding, default to utf-8
            encoding = "utf-8"

        return encoding

    def _get_default_timeout(self) -> Optional[int]:
        return None

    def _get_impl_class(self) -> "Type[SubprocessImpl]":
        """
        Return the ``impl`` class to use.
        """
        return SubprocessImpl

    def __attrs_post_init__(self) -> None:
        """
        Post ``attrs`` class initialization routines.
        """
        impl_class = self._get_impl_class()
        self.impl = impl_class(factory=self)

    def cmdline(self, *args: str) -> List[str]:
        """
        Method to construct a command line.
        """
        self._cmdline = list(args)
        return self._cmdline

    def get_display_name(self) -> str:
        """
        Returns a human readable name for the factory.
        """
        return "{}({})".format(self.__class__.__name__, self._cmdline or "")

    def is_running(self) -> bool:
        """
        Returns true if the sub-process is alive.
        """
        return self.impl.is_running()

    def terminate(self) -> ProcessResult:
        """
        Terminate the started subprocess.
        """
        return self.impl.terminate()

    @property
    def pid(self) -> Optional[int]:
        """
        The pid of the running process. None if not running.
        """
        return self.impl.pid


@attr.s(slots=True, kw_only=True)
class Subprocess(Factory):
    """
    Base shell factory class.
    """

    def run(
        self,
        *args: str,
        env: Optional[EnvironDict] = None,
        _timeout: Optional[Union[int, float]] = None,
        **kwargs: Any,
    ) -> ProcessResult:
        """
        Run the given command synchronously.

        Keyword Arguments:
            args:
                The list of arguments to pass to :py:meth:`~pytestshellutils.shell.Subprocess.cmdline`
                to construct the command to run
            env:
                Pass a dictionary of environment key, value pairs to inject into the subprocess.
            _timeout:
                The timeout value for this particular ``run()`` call. If this value is not ``None``,
                it will be used instead of :py:attr:`~pytestshellutils.shell.Subprocess.timeout`,
                the default timeout.
        """
        start_time = time.time()
        # Build the cmdline to pass to the terminal
        # We set the _terminal_timeout attribute while calling cmdline in case it needs
        # access to that information to build the command line
        self.impl._terminal_timeout = _timeout or self.timeout
        timmed_out = False
        try:
            self.impl.run(*args, env=env, **kwargs)
            if TYPE_CHECKING:
                # Make mypy happy
                assert self.impl._terminal
            self.impl._terminal.communicate(timeout=self.impl._terminal_timeout)
        except subprocess.TimeoutExpired:
            timmed_out = True

        result = self.terminate()
        cmdline = result.cmdline
        returncode = result.returncode
        if timmed_out:
            raise FactoryTimeout(
                "{} Failed to run: {}; Error: Timed out after {:.2f} seconds!".format(
                    self, cmdline, time.time() - start_time
                ),
                process_result=result,
            )
        stdout, stderr, json_out = self.process_output(
            result.stdout, result.stderr, cmdline=cmdline
        )
        log.info(
            "%s completed %r in CWD: %s after %.2f seconds",
            self,
            cmdline,
            self.cwd,
            time.time() - start_time,
        )
        return ProcessResult(
            returncode=returncode, stdout=stdout, stderr=stderr, data=json_out, cmdline=cmdline
        )

    def process_output(
        self, stdout: str, stderr: str, cmdline: Optional[List[str]] = None
    ) -> Tuple[str, str, Optional[Dict[Any, Any]]]:
        """
        Process the output. When possible JSON is loaded from the output.

        Returns:
            Returns a tuple in the form of ``(stdout, stderr, loaded_json)``
        """
        if stdout:
            try:
                json_out = json.loads(stdout)
            except ValueError:
                log.debug("%s failed to load JSON from the following output:\n%r", self, stdout)
                json_out = None
        else:
            json_out = None
        return stdout, stderr, json_out


@attr.s(slots=True, kw_only=True)
class ScriptSubprocess(Subprocess):
    """
    Base CLI script/binary class.

    Keyword Arguments:
        script_name:
            This is the string containing the name of the binary to call on the subprocess, either the
            full path to it, or the basename. In case of the basename, the directory containing the
            basename must be in your ``$PATH`` variable.
        base_script_args:
            An list or tuple iterable of the base arguments to use when building the command line to
            launch the process

    Please look at :py:class:`~pytestshellutils.shell.Factory` for the additional supported keyword
    arguments documentation.
    """

    script_name: str = attr.ib()
    base_script_args: List[str] = attr.ib(factory=list)

    def get_display_name(self) -> str:
        """
        Returns a human readable name for the factory.
        """
        return f"{self.__class__.__name__}({pathlib.Path(self.script_name).name})"

    def get_script_path(self) -> str:
        """
        Returns the path to the script to run.
        """
        script_path: Optional[str]
        if os.path.isabs(self.script_name):
            script_path = self.script_name
        else:
            script_path = shutil.which(self.script_name)
        if not script_path or not os.path.exists(script_path):
            raise FileNotFoundError(f"The CLI script '{self.script_name}' does not exist")
        if TYPE_CHECKING:
            # Make mypy happy
            assert script_path
        return script_path

    def get_base_script_args(self) -> List[str]:
        """
        Returns any additional arguments to pass to the CLI script.
        """
        return list(self.base_script_args)

    def get_script_args(self) -> List[str]:  # pylint: disable=no-self-use
        """
        Returns any additional arguments to pass to the CLI script.
        """
        return []

    def cmdline(self, *args: str) -> List[str]:
        """
        Construct a list of arguments to use when starting the subprocess.

        Arguments:
            args:
                Additional arguments to use when starting the subprocess
        """
        return (
            [self.get_script_path()]
            + self.get_base_script_args()
            + self.get_script_args()
            + list(args)
        )


@attr.s(kw_only=True, slots=True, frozen=True)
class StartDaemonCallArguments:
    """
    This class holds the arguments and keyword arguments used to start a daemon.

    It's used when restarting the daemon so that the same call is used.

    Keyword Arguments:
        args:
            List of arguments
        kwargs:
            Dictionary of keyword arguments
    """

    args: Tuple[str, ...] = attr.ib()
    kwargs: Dict[str, Any] = attr.ib()


@attr.s(slots=True, kw_only=True)
class DaemonImpl(SubprocessImpl):
    """
    Daemon subprocess interaction implementation.

    Please look at :py:class:`~pytestshellutils.shell.SubprocessImpl` for the additional supported keyword
    arguments documentation.
    """

    factory: "Daemon" = attr.ib()

    _before_start_callbacks: List[Callback] = attr.ib(repr=False, hash=False, factory=list)
    _after_start_callbacks: List[Callback] = attr.ib(repr=False, hash=False, factory=list)
    _before_terminate_callbacks: List[Callback] = attr.ib(repr=False, hash=False, factory=list)
    _after_terminate_callbacks: List[Callback] = attr.ib(repr=False, hash=False, factory=list)
    _start_args_and_kwargs: StartDaemonCallArguments = attr.ib(
        init=False, repr=False, hash=False, default=None
    )

    def before_start(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run before the daemon starts.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self._before_start_callbacks.append(Callback(func=callback, args=args, kwargs=kwargs))

    def after_start(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run after the daemon starts.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self._after_start_callbacks.append(Callback(func=callback, args=args, kwargs=kwargs))

    def before_terminate(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run before the daemon terminates.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self._before_terminate_callbacks.append(Callback(func=callback, args=args, kwargs=kwargs))

    def after_terminate(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run after the daemon terminates.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self._after_terminate_callbacks.append(Callback(func=callback, args=args, kwargs=kwargs))

    def start(
        self,
        *extra_cli_arguments: str,
        max_start_attempts: Optional[int] = None,
        start_timeout: Optional[Union[int, float]] = None,
    ) -> bool:
        """
        Start the daemon.

        Keyword Arguments:
            extra_cli_arguments:
                Extra arguments to pass to the CLI that starts the daemon
            max_start_attempts:
                Maximum number of attempts to try and start the daemon in case of failures
            start_timeout:
                The maximum number of seconds to wait before considering that the daemon did not start

        Returns:
            bool: A boolean indicating if the start was successful or not.
        """
        if self.is_running():  # pragma: no cover
            log.warning("%s is already running.", self)
            return True
        self._start_args_and_kwargs = StartDaemonCallArguments(
            args=extra_cli_arguments,
            kwargs={"max_start_attempts": max_start_attempts, "start_timeout": start_timeout},
        )
        process_running = False
        start_time = time.time()
        start_attempts = max_start_attempts or self.factory.max_start_attempts
        current_attempt = 0
        run_arguments = list(extra_cli_arguments)
        while True:
            if process_running:
                break
            current_attempt += 1
            if current_attempt > start_attempts:
                break
            log.info(
                "Starting %s. Attempt: %d of %d", self.factory, current_attempt, start_attempts
            )
            for callback in self._before_start_callbacks:  # pylint: disable=not-an-iterable
                try:
                    callback()
                except CallbackException as exc:  # pragma: no cover
                    log.info(
                        "Exception raised when running %s: %s",
                        callback,
                        exc,
                        exc_info=True,
                    )
            current_start_time = time.time()
            start_running_timeout = current_start_time + (
                start_timeout or self.factory.start_timeout
            )
            if current_attempt > 1 and self.factory.extra_cli_arguments_after_first_start_failure:
                run_arguments = list(extra_cli_arguments) + list(
                    self.factory.extra_cli_arguments_after_first_start_failure
                )
            self.run(*run_arguments)
            if not self.is_running():  # pragma: no cover
                # A little breathe time to allow the process to start if not started already
                time.sleep(0.5)
            while time.time() <= start_running_timeout:
                if not self.is_running():
                    log.warning("%s is no longer running", self.factory)
                    self.terminate()
                    break
                try:
                    if (
                        self.factory.run_start_checks(current_start_time, start_running_timeout)
                        is False
                    ):
                        time.sleep(1)
                        continue
                except FactoryNotStarted:
                    self.terminate()
                    break
                log.info(
                    "The %s factory is running after %d attempts. Took %1.2f seconds",
                    self.factory,
                    current_attempt,
                    time.time() - start_time,
                )
                process_running = True
                break
            else:
                # The factory failed to confirm it's running status
                self.terminate()
        if process_running:
            for callback in self._after_start_callbacks:  # pylint: disable=not-an-iterable
                try:
                    callback()
                except CallbackException as exc:  # pragma: no cover
                    log.info(
                        "Exception raised when running %s: %s",
                        callback,
                        exc,
                        exc_info=True,
                    )
            return process_running
        result = self.terminate()
        raise FactoryNotStarted(
            "The {} factory has failed to confirm running status after {} attempts, which "
            "took {:.2f} seconds".format(
                self.factory,
                current_attempt - 1,
                time.time() - start_time,
            ),
            process_result=result,
        )

    def terminate(self) -> ProcessResult:
        """
        Terminate the daemon.
        """
        if self._terminal_result is not None:
            # This factory has already been terminated
            return self._terminal_result
        for callback in self._before_terminate_callbacks:  # pylint: disable=not-an-iterable
            try:
                callback()
            except CallbackException as exc:  # pragma: no cover
                log.info(
                    "Exception raised when running %s: %s",
                    callback,
                    exc,
                    exc_info=True,
                )
        try:
            return super().terminate()
        finally:
            for callback in self._after_terminate_callbacks:  # pylint: disable=not-an-iterable
                try:
                    callback()
                except CallbackException as exc:  # pragma: no cover
                    log.warning(
                        "Exception raised when running %s: %s",
                        callback,
                        exc,
                        exc_info=True,
                    )

    def get_start_arguments(self) -> StartDaemonCallArguments:
        """
        Return the arguments and keyword arguments used when starting the daemon.
        """
        return self._start_args_and_kwargs


@attr.s(slots=True, kw_only=True)
class Daemon(ScriptSubprocess):
    """
    Base daemon factory.

    Keyword Arguments:
        check_ports:
            List of ports to try and connect to while confirming that the daemon is up and running
        extra_cli_arguments_after_first_start_failure:
            Extra arguments to pass to the CLI that starts the daemon after the first failure
        max_start_attempts:
            Maximum number of attempts to try and start the daemon in case of failures
        start_timeout:
            The maximum number of seconds to wait before considering that the daemon did not start

    Please look at :py:class:`~pytestshellutils.shell.Subprocess` for the additional supported keyword
    arguments documentation.
    """

    impl: DaemonImpl = attr.ib(repr=False, init=False)

    script_name: str = attr.ib()
    base_script_args: List[str] = attr.ib(factory=list)
    check_ports: List[int] = attr.ib(factory=list)
    stats_processes: "StatsProcesses" = attr.ib(repr=False, hash=False, default=None)
    start_timeout: Union[int, float] = attr.ib(repr=False)
    max_start_attempts: int = attr.ib(repr=False, default=3)
    extra_cli_arguments_after_first_start_failure: List[str] = attr.ib(hash=False, factory=list)
    listen_ports: List[int] = attr.ib(init=False, repr=False, hash=False, factory=list)
    _start_checks_callbacks: List[Callback] = attr.ib(repr=False, hash=False, factory=list)

    def _get_impl_class(self) -> "Type[DaemonImpl]":
        """
        Return the ``impl`` class to use.
        """
        return DaemonImpl

    def __attrs_post_init__(self) -> None:
        """
        Post ``attrs`` class initialization routines.
        """
        super().__attrs_post_init__()
        if self.check_ports and not isinstance(self.check_ports, (list, tuple)):
            self.check_ports = [self.check_ports]
        if self.check_ports:
            self.listen_ports.extend(self.check_ports)

        self.after_start(self._add_factory_to_stats_processes)
        self.after_terminate(self._terminate_processes_matching_listen_ports)
        self.after_terminate(self._remove_factory_from_stats_processes)
        self.start_check(self._check_listening_ports)

    def before_start(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run before the daemon starts.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self.impl.before_start(callback, *args, **kwargs)

    def after_start(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run after the daemon starts.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self.impl.after_start(callback, *args, **kwargs)

    def before_terminate(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run before the daemon terminates.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self.impl.before_terminate(callback, *args, **kwargs)

    def after_terminate(self, callback: Callable[[], None], *args: Any, **kwargs: Any) -> None:
        """
        Register a function callback to run after the daemon terminates.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.
        """
        self.impl.after_terminate(callback, *args, **kwargs)

    def start_check(self, callback: Callable[..., bool], *args: Any, **kwargs: Any) -> None:
        """
        Register a function to run after the daemon starts to confirm readiness for work.

        The callback must accept as the first argument ``timeout_at`` which is a float.
        The callback must stop trying to confirm running behavior once ``time.time() > timeout_at``.
        The callback should return ``True`` to confirm that the daemon is ready for work.

        Arguments:
            callback:
                The function to call back

        Keyword Arguments:
            args:
                The arguments to pass to the callback
            kwargs:
                The keyword arguments to pass to the callback

        Returns:
            Nothing.

        Example:
            .. code-block:: python

                def check_running_state(timeout_at: float) -> bool:
                    while time.time() <= timeout_at:
                        # run some checks
                        ...
                        # if all is good
                        break
                    else:
                        return False
                    return True
        """
        self._start_checks_callbacks.append(Callback(func=callback, args=args, kwargs=kwargs))

    def get_check_ports(self) -> List[int]:
        """
        Return a list of ports to check against to ensure the daemon is running.
        """
        return self.check_ports or []

    def get_start_check_callbacks(self) -> List[Callback]:
        """
        Return a list of the start check callbacks.
        """
        return self._start_checks_callbacks or []

    def start(
        self,
        *extra_cli_arguments: str,
        max_start_attempts: Optional[int] = None,
        start_timeout: Optional[Union[int, float]] = None,
    ) -> bool:
        """
        Start the daemon.
        """
        return self.impl.start(
            *extra_cli_arguments, max_start_attempts=max_start_attempts, start_timeout=start_timeout
        )

    @contextlib.contextmanager
    def started(
        self,
        *extra_cli_arguments: str,
        max_start_attempts: Optional[int] = None,
        start_timeout: Optional[Union[int, float]] = None,
    ) -> Generator["Daemon", None, None]:
        """
        Start the daemon and return it's instance so it can be used as a context manager.
        """
        try:
            self.start(
                *extra_cli_arguments,
                max_start_attempts=max_start_attempts,
                start_timeout=start_timeout,
            )
            yield self
        finally:
            self.terminate()

    @contextlib.contextmanager
    def stopped(
        self,
        before_stop_callback: Optional[Callable[["Daemon"], None]] = None,
        after_stop_callback: Optional[Callable[["Daemon"], None]] = None,
        before_start_callback: Optional[Callable[["Daemon"], None]] = None,
        after_start_callback: Optional[Callable[["Daemon"], None]] = None,
    ) -> Generator["Daemon", None, None]:
        """
        Stop the daemon and return it's instance so it can be used as a context manager.

        Keyword Arguments:
            before_stop_callback:
                A callable to run before stopping the daemon. The callback must accept one argument,
                the daemon instance.
            after_stop_callback:
                A callable to run after stopping the daemon. The callback must accept one argument,
                the daemon instance.
            before_start_callback:
                A callable to run before starting the daemon. The callback must accept one argument,
                the daemon instance.
            after_start_callback:
                A callable to run after starting the daemon. The callback must accept one argument,
                the daemon instance.

        This context manager will stop the factory while the context is in place, it re-starts it once out of
        context.

        Example:
            .. code-block:: python

                assert factory.is_running() is True

                with factory.stopped():
                    assert factory.is_running() is False

                assert factory.is_running() is True
        """
        if not self.is_running():
            raise FactoryNotRunning(f"{self} is not running ")
        start_arguments = self.impl.get_start_arguments()
        try:
            if before_stop_callback:
                try:
                    before_stop_callback(self)
                except CallbackException as exc:  # pragma: no cover
                    log.info(
                        "Exception raised when running %s: %s",
                        format_callback_to_string(before_stop_callback),
                        exc,
                        exc_info=True,
                    )
            self.terminate()
            if after_stop_callback:
                try:
                    after_stop_callback(self)
                except CallbackException as exc:  # pragma: no cover
                    log.info(
                        "Exception raised when running %s: %s",
                        format_callback_to_string(after_stop_callback),
                        exc,
                        exc_info=True,
                    )
            yield self
        except ShellUtilsException:  # pragma: no cover pylint: disable=try-except-raise
            raise
        else:
            if before_start_callback:
                try:
                    before_start_callback(self)
                except CallbackException as exc:  # pragma: no cover
                    log.info(
                        "Exception raised when running %s: %s",
                        format_callback_to_string(before_start_callback),
                        exc,
                        exc_info=True,
                    )
            _started = self.start(
                *start_arguments.args,  # pylint: disable=not-an-iterable
                **start_arguments.kwargs,  # pylint: disable=not-a-mapping
            )
            if _started:
                if after_start_callback:
                    try:
                        after_start_callback(self)
                    except CallbackException as exc:  # pragma: no cover
                        log.info(
                            "Exception raised when running %s: %s",
                            format_callback_to_string(after_start_callback),
                            exc,
                            exc_info=True,
                        )

    def run_start_checks(self, started_at: float, timeout_at: float) -> bool:
        """
        Run checks to confirm that the daemon has started.
        """
        start_check_callbacks = list(self.get_start_check_callbacks())
        if not start_check_callbacks:
            log.debug("No start check callbacks to run for %s", self)
            return True
        checks_start_time = time.time()
        log.debug("%s is running start checks", self)
        while time.time() <= timeout_at:
            if not self.is_running():
                raise FactoryNotStarted(f"{self} is no longer running")
            if not start_check_callbacks:
                break
            start_check = start_check_callbacks[0]
            try:
                ret = start_check(timeout_at)
                if ret is True:
                    start_check_callbacks.pop(0)
            except Exception as exc:  # pylint: disable=broad-except
                log.info(
                    "Exception raised when running %s: %s",
                    start_check,
                    exc,
                    exc_info=True,
                )
        if start_check_callbacks:
            log.error(
                "Failed to run start check callbacks after %1.2f seconds for %s. "
                "Remaining start check callbacks: %s",
                time.time() - checks_start_time,
                self,
                start_check_callbacks,
            )
            return False
        log.debug("All start check callbacks executed for %s", self)
        return True

    def _check_listening_ports(self, timeout_at: float) -> bool:
        """
        Check if the defined ports are in a listening state.

        This callback will run when trying to assess if the daemon is ready
        to accept work by trying to connect to each of the ports it's supposed
        to be listening.
        """
        check_ports = set(self.get_check_ports())
        if not check_ports:
            log.debug("No ports to check connection to for %s", self)
            return True
        log.debug("Listening ports to check for %s: %s", self, set(self.get_check_ports()))
        checks_start_time = time.time()
        while time.time() <= timeout_at:
            if not self.is_running():
                raise FactoryNotStarted(f"{self} is no longer running")
            if not check_ports:
                break
            check_ports -= ports.get_connectable_ports(check_ports)
            if check_ports:
                time.sleep(1.5)
        else:
            log.error(
                "Failed to check ports after %1.2f seconds for %s. Remaining ports to check: %s",
                time.time() - checks_start_time,
                self,
                check_ports,
            )
            return False
        log.debug("All listening ports checked for %s: %s", self, set(self.get_check_ports()))
        return True

    def _add_factory_to_stats_processes(self) -> None:
        if self.stats_processes is not None:
            display_name = self.get_display_name()
            self.stats_processes.add(display_name, self.pid)

    def _remove_factory_from_stats_processes(self) -> None:
        if self.stats_processes is not None:
            display_name = self.get_display_name()
            self.stats_processes.remove(display_name)

    def _terminate_processes_matching_listen_ports(self) -> None:
        if not self.listen_ports:
            return
        # If any processes were not terminated and are listening on the ports
        # we have set on listen_ports, terminate those processes.
        found_processes = []
        psutil_majorver, _, _ = psutil.version_info
        if psutil_majorver < 6:
            for process in psutil.process_iter(["connections"]):
                try:
                    for connection in process.connections():
                        if connection.status != psutil.CONN_LISTEN:
                            # We only care about listening services
                            continue
                        if connection.laddr.port in self.check_ports:
                            found_processes.append(process)
                            # We already found one connection, no need to check the others
                            break
                except psutil.AccessDenied:  # pragma: no cover
                    # We've been denied access to this process connections. Carry on.
                    continue
                except psutil.ZombieProcess:
                    continue
        else:
            for process in psutil.process_iter(["net_connections"]):
                try:
                    for connection in process.net_connections():
                        if connection.status != psutil.CONN_LISTEN:
                            # We only care about listening services
                            continue
                        if connection.laddr.port in self.check_ports:
                            found_processes.append(process)
                            # We already found one connection, no need to check the others
                            break
                except psutil.AccessDenied:  # pragma: no cover
                    # We've been denied access to this process net_connections. Carry on.
                    continue
                except psutil.ZombieProcess:
                    continue
        if found_processes:
            log.debug(
                "The following processes were found listening on ports %s: %s",
                ", ".join(
                    [str(port) for port in self.listen_ports],  # pylint: disable=not-an-iterable
                ),
                found_processes,
            )
            terminate_process_list(found_processes, kill=True, slow_stop=False)
        else:
            log.debug(
                "No astray processes were found listening on ports: %s",
                ", ".join(
                    [str(port) for port in self.listen_ports],  # pylint: disable=not-an-iterable
                ),
            )

    def __enter__(self) -> "Daemon":
        """
        Use class as a context manager.
        """
        if not self.is_running():
            raise RuntimeError(
                "Factory not yet started. Perhaps you're after something like:\n\n"
                "with {}.started() as factory:\n"
                "    yield factory".format(self.__class__.__name__)
            )
        return self

    def __exit__(self, *_: Any) -> None:
        """
        Exit the class context manager.
        """
        self.terminate()



#!/usr/bin/env python3
# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=invalid-name,missing-module-docstring,missing-function-docstring
import argparse
import pathlib
import re
import sys

CODE_ROOT = pathlib.Path(__file__).resolve().parent.parent
CHANGELOG_ENTRIES_PATH = CODE_ROOT / "changelog"
CHANGELOG_LIKE_RE = re.compile(r"([\d]+)\.([a-z]+)(\.rst)?$")
CHANGELOG_EXTENSIONS = (
    "breaking",
    "deprecation",
    "feature",
    "improvement",
    "bugfix",
    "doc",
    "trivial",
)
CHANGELOG_ENTRY_REREX = r"^[\d]+\.({})\.rst$".format("|".join(CHANGELOG_EXTENSIONS))
CHANGELOG_ENTRY_RE = re.compile(CHANGELOG_ENTRY_REREX)


def check_changelog_entries(files):

    exitcode = 0
    for entry in files:
        path = pathlib.Path(entry).resolve()
        # Is it under changelog/
        try:
            path.relative_to(CHANGELOG_ENTRIES_PATH)
            if path.name in (".gitkeep", ".gitignore", "_template.rst", __name__):
                # These files should be ignored
                continue
            # Is it named properly
            if not CHANGELOG_ENTRY_RE.match(path.name):
                # Does it end in .rst
                if path.suffix != ".rst":
                    exitcode = 1
                    print(
                        "The changelog entry '{}' should have '.rst' as it's file extension".format(
                            path.relative_to(CODE_ROOT),
                        ),
                        file=sys.stderr,
                        flush=True,
                    )
                    continue
                print(
                    "The changelog entry '{}' should have one of the following extensions: {}.".format(
                        path.relative_to(CODE_ROOT),
                        ", ".join(repr(ext) for ext in CHANGELOG_EXTENSIONS),
                    ),
                    file=sys.stderr,
                    flush=True,
                )
                exitcode = 1
                continue
            check_changelog_entry_contents(path)
        except ValueError:
            # Not under changelog/, carry on checking
            # Is it a changelog entry
            if CHANGELOG_ENTRY_RE.match(path.name):
                # So, this IS a changelog entry, but it's misplaced....
                exitcode = 1
                print(
                    "The changelog entry '{}' should be placed under '{}/', not '{}'".format(
                        path.relative_to(CODE_ROOT),
                        CHANGELOG_ENTRIES_PATH.relative_to(CODE_ROOT),
                        path.relative_to(CODE_ROOT).parent,
                    ),
                    file=sys.stderr,
                    flush=True,
                )
                continue
            elif CHANGELOG_LIKE_RE.match(path.name) and not CHANGELOG_ENTRY_RE.match(path.name):
                # Does it look like a changelog entry
                print(
                    "The changelog entry '{}' should have one of the following extensions: {}.".format(
                        path.relative_to(CODE_ROOT),
                        ", ".join(repr(ext) for ext in CHANGELOG_EXTENSIONS),
                    ),
                    file=sys.stderr,
                    flush=True,
                )
                exitcode = 1
                continue

            elif not CHANGELOG_LIKE_RE.match(path.name) and not CHANGELOG_ENTRY_RE.match(path.name):
                # Does not look like, and it's not a changelog entry
                continue
            # Does it end in .rst
            if path.suffix != ".rst":
                exitcode = 1
                print(
                    "The changelog entry '{}' should have '.rst' as it's file extension".format(
                        path.relative_to(CODE_ROOT),
                    ),
                    file=sys.stderr,
                    flush=True,
                )
    return exitcode


def check_changelog_entry_contents(entry):
    contents = entry.read_text().splitlines()
    if len(contents) > 1:
        # More than one line.
        # If the second line starts with '*' it's a bullet list and we need to add an
        # empty line before it.
        if contents[1].strip().startswith("*"):
            contents.insert(1, "")
    entry.write_text("{}\n".format("\n".join(contents)))


def main(argv):
    parser = argparse.ArgumentParser(prog=__name__)
    parser.add_argument("files", nargs="+")

    options = parser.parse_args(argv)
    return check_changelog_entries(options.files)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Custom Types.
"""
import copy
import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import Protocol
from typing import Tuple
from typing import TYPE_CHECKING

import attr

from pytestshellutils.utils import format_callback_to_string

if TYPE_CHECKING:
    from pytestshellutils.shell import Daemon


log = logging.getLogger(__name__)


class EnvironDict(Dict[str, str]):
    """
    Environ dictionary type.
    """

    def __str__(self) -> str:  # pragma: no cover
        """
        String representation of the class.
        """
        return f"EnvironDict({super().__str__()})"


class GenericCallback(Protocol):
    """
    Generic callback function.
    """

    def __call__(self, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
        """
        Call the generic callback.
        """
        ...


class DaemonCallback(Protocol):
    """
    Daemon callback function.
    """

    def __call__(self, daemon: "Daemon") -> None:  # pragma: no cover
        """
        Call the daemon callback.
        """
        ...


@attr.s(kw_only=True, frozen=True)
class Callback:
    """
    Class which "stores" information of a callback.
    """

    func: Callable[..., Any] = attr.ib()
    args: Tuple[Any, ...] = attr.ib(default=None)
    kwargs: Dict[str, Any] = attr.ib(default=None)

    def __str__(self) -> str:
        """
        String representation of the class.
        """
        return format_callback_to_string(self.func, self.args, self.kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Call the callback.
        """
        _args = tuple(list(args) + list(self.args or ()))
        _kwargs = copy.deepcopy(self.kwargs)
        _kwargs.update(kwargs)
        log.debug("Running %s", format_callback_to_string(self.func, _args, _kwargs))
        return self.func(*_args, **_kwargs)



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=wildcard-import,unused-wildcard-import
"""
Namespace the standard library :py:mod:`socket` module.

This module's sole purpose is to have the standard library :py:mod:`socket` module functions under a different
namespace to be used in pytest-shell-utilities so that projects using it, which need to mock :py:mod:`socket` functions,
don't influence the pytest-shell-utilities run time behavior.
"""
from socket import *



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import functools
import logging
import os
import stat
import tempfile
import textwrap
from typing import Optional
from typing import Tuple

import pytest

try:
    from pytest import FixtureRequest
except ImportError:
    from _pytest.fixtures import FixtureRequest

try:  # pragma: no cover
    import importlib.metadata

    pkg_version = importlib.metadata.version
except ImportError:  # pragma: no cover
    try:
        import importlib_metadata

        pkg_version = importlib_metadata.version
    except ImportError:  # pragma: no cover
        import pkg_resources

        def pkg_version(package):  # type: ignore[no-untyped-def]
            return pkg_resources.get_distribution(package).version


log = logging.getLogger(__name__)


def pkg_version_info(package: str) -> Tuple[int, ...]:
    """
    Return a version info tuple for the given package.
    """
    return tuple(int(part) for part in pkg_version(package).split(".") if part.isdigit())


if pkg_version_info("pytest") >= (6, 2):
    pytest_plugins = ["pytester"]
else:  # pragma: no cover

    @pytest.fixture
    def pytester() -> None:
        pytest.skip("The pytester fixture is not available in Pytest < 6.2.0")


class Tempfiles:
    """
    Class which generates temporary files and cleans them when done.
    """

    def __init__(self, request: FixtureRequest):
        self.request = request

    def makepyfile(
        self, contents: str, prefix: Optional[str] = None, executable: bool = False
    ) -> str:
        """
        Creates a python file and returns it's path.
        """
        tfile = tempfile.NamedTemporaryFile("w", prefix=prefix or "tmp", suffix=".py", delete=False)
        contents = textwrap.dedent(contents.lstrip("\n")).strip()
        tfile.write(contents)
        tfile.close()
        if executable is True:
            st = os.stat(tfile.name)
            os.chmod(tfile.name, st.st_mode | stat.S_IEXEC)
        self.request.addfinalizer(functools.partial(self._delete_temp_file, tfile.name))
        with open(tfile.name, encoding="utf-8") as rfh:
            log.debug(
                "Created python file with contents:\n>>>>> %s >>>>>\n%s\n<<<<< %s <<<<<\n",
                tfile.name,
                rfh.read(),
                tfile.name,
            )
        return tfile.name

    def _delete_temp_file(self, fpath: str) -> None:
        """
        Cleanup the temporary path.
        """
        if os.path.exists(fpath):  # pragma: no branch
            os.unlink(fpath)


@pytest.fixture
def tempfiles(request: FixtureRequest) -> Tempfiles:
    """
    Temporary files fixture.
    """
    return Tempfiles(request)



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Process related utilities.
"""
import errno
import json
import logging
import pprint
import signal
import weakref
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import attr
import psutil

try:
    from pytest import LineMatcher
except ImportError:  # pragma: no cover
    # Older pytest
    from _pytest.pytester import LineMatcher

from pytestshellutils.utils import warn_until


log = logging.getLogger(__name__)


class MatchString(str):
    """
    Simple subclass around ``str`` which provides a ``.matcher`` property.

    This ``.matcher`` property is an instance of :py:class:`~pytest.LineMatcher`
    """

    @property
    def matcher(self) -> LineMatcher:
        """
        Return an instance of :py:class:`~pytest.LineMatcher`.
        """
        return LineMatcher(self.splitlines())


def convert_string_to_match_string(value: Optional[str]) -> Optional[MatchString]:
    """
    Convert strings into ``MatchString`` instances.
    """
    if isinstance(value, str):
        return MatchString(value)
    return value


@attr.s(frozen=True, kw_only=True)
class ProcessResult:
    """
    Wrapper class around a subprocess result.

    This class serves the purpose of having a common result class which will hold the
    resulting data from a subprocess command.

    Keyword Arguments:
        returncode:
            The returncode returned by the process
        stdout:
            The ``stdout`` returned by the process
        stderr:
            The ``stderr`` returned by the process
        cmdline:
            The command line used to start the process
        data:
            The data returned by parsing ``stdout``, when possible.
        data_key:
            When ``stdout`` can be parsed as JSON, sometimes there's a top level key which is not
            that interesting. By using ``data_key``, we define that we're actually only interested
            on the data structure which is keyed by ``data_key``.

    Note:
        Cast :py:class:`~pytestshellutils.utils.processes.ProcessResult` to a string to pretty-print it.
    """

    returncode: int = attr.ib()
    stdout: MatchString = attr.ib(converter=convert_string_to_match_string)
    stderr: MatchString = attr.ib(converter=convert_string_to_match_string)
    cmdline: Optional[List[str]] = attr.ib(default=None)
    data_key: Optional[str] = attr.ib(default=None)
    data: Optional[Dict[Any, Any]] = attr.ib()

    @returncode.validator
    def _validate_returncode(self, attribute: Any, value: int) -> None:
        """
        Validate the value type.
        """
        if not isinstance(value, int):
            raise ValueError(f"'returncode' needs to be an integer, not '{type(value)}'")

    @data.default
    def _default_data(self) -> Optional[Dict[Any, Any]]:
        """
        Try to parse the passed ``stdout`` as JSON as the default data value.
        """
        stdout: Optional[str] = self.stdout.strip() if self.stdout else None
        if stdout:
            try:
                data: Optional[Dict[Any, Any]] = json.loads(stdout.strip())
                if data and self.data_key and self.data_key in data:
                    data = data[self.data_key]
                return data
            except ValueError:
                pass
        return None

    @property
    def exitcode(self) -> int:  # pragma: no cover
        """
        Return the process returncode.

        This property is deprecated and should not be used.
        It only exists to support projects that are migrating from
        pytest-salt-factories versions. Use ``.returncode`` instead.
        """
        warn_until(
            "2.0.0",
            "The '.exitcode' property is deprecated and will cease to exist after "
            "pytest-shell-utilities {version}. Please use '.returncode' instead.",
        )
        return self.returncode

    @property
    def json(self) -> Optional[Dict[Any, Any]]:  # pragma: no cover
        """
        Return the process output parsed as JSON, if possible.

        This property is deprecated and should not be used.
        It only exists to support projects that are migrating from
        pytest-salt-factories versions. Use ``.data`` instead.
        """
        warn_until(
            "2.0.0",
            "The '.json' property is deprecated and will cease to exist after "
            "pytest-shell-utilities {version}. Please use '.data' instead.",
        )
        return self.data

    def __str__(self) -> str:
        """
        String representation of the class.
        """
        message = self.__class__.__name__
        if self.cmdline:
            message += f"\n Command Line: {self.cmdline}"
        if self.returncode is not None:
            message += f"\n Returncode: {self.returncode}"
        if (self.stdout and self.stdout.strip()) or (self.stderr and self.stderr.strip()):
            message += "\n Process Output:"
        if self.stdout and self.stdout.strip():
            message += f"\n   >>>>> STDOUT >>>>>\n{self.stdout}\n   <<<<< STDOUT <<<<<"
        if self.stderr and self.stderr.strip():
            message += f"\n   >>>>> STDERR >>>>>\n{self.stderr}\n   <<<<< STDERR <<<<<"
        if self.data:
            message += f"\n Parsed JSON Data:\n{self._to_printable_data(self.data)}"
        return message + "\n"

    @staticmethod
    def _to_printable_data(data: Dict[str, Any]) -> str:
        """
        Convert a data dictionary into a JSON string.
        """
        return "\n".join(
            f"   {line}" for line in json.dumps(data, indent=2, sort_keys=False).splitlines()
        ).rstrip()


def collect_child_processes(pid: int) -> List[psutil.Process]:
    """
    Try to collect any started child processes of the provided pid.

    Arguments:
        pid:
            The PID of the process

    Returns:
        List of child processes
    """
    # Let's get the child processes of the started subprocess
    children: List[psutil.Process]
    try:
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
    except psutil.NoSuchProcess:
        children = []
    return children


def _get_cmdline(proc: psutil.Process) -> Optional[Any]:
    # pylint: disable=protected-access
    try:
        return proc._cmdline
    except AttributeError:
        # Cache the cmdline since that will be inaccessible once the process is terminated
        # and we use it in log calls
        try:
            cmdline = proc.cmdline()
        except (psutil.NoSuchProcess, psutil.AccessDenied):  # pragma: no cover
            # OSX is more restrictive about the above information
            cmdline = None
        except OSError:  # pragma: no cover
            # On Windows we've seen something like:
            #   File " c: ... \lib\site-packages\pytestsalt\utils\__init__.py", line 182, in terminate_process
            #     terminate_process_list(process_list, kill=slow_stop is False, slow_stop=slow_stop)
            #   File " c: ... \lib\site-packages\pytestsalt\utils\__init__.py", line 130, in terminate_process_list
            #     _terminate_process_list(process_list, kill=kill, slow_stop=slow_stop)
            #   File " c: ... \lib\site-packages\pytestsalt\utils\__init__.py", line 78, in _terminate_process_list
            #     cmdline = process.cmdline()
            #   File " c: ... \lib\site-packages\psutil\__init__.py", line 786, in cmdline
            #     return self._proc.cmdline()
            #   File " c: ... \lib\site-packages\psutil\_pswindows.py", line 667, in wrapper
            #     return fun(self, *args, **kwargs)
            #   File " c: ... \lib\site-packages\psutil\_pswindows.py", line 745, in cmdline
            #     ret = cext.proc_cmdline(self.pid, use_peb=True)
            #   OSError: [WinError 299] Only part of a ReadProcessMemory or WriteProcessMemory request was completed: 'originated from ReadProcessMemory(ProcessParameters)
            cmdline = None
        except RuntimeError:  # pragma: no cover
            # Also on windows
            # saltfactories\utils\processes\helpers.py:68: in _get_cmdline
            #     cmdline = proc.as_dict()
            # c: ... \lib\site-packages\psutil\__init__.py:634: in as_dict
            #     ret = meth()
            # c: ... \lib\site-packages\psutil\__init__.py:1186: in memory_full_info
            #     return self._proc.memory_full_info()
            # c: ... \lib\site-packages\psutil\_pswindows.py:667: in wrapper
            #     return fun(self, *args, **kwargs)
            # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
            #
            # self = <psutil._pswindows.Process object at 0x0000029B7FDA5558>
            #
            #     @wrap_exceptions
            #     def memory_full_info(self):
            #         basic_mem = self.memory_info()
            # >       uss = cext.proc_memory_uss(self.pid)
            # E       RuntimeError: NtQueryVirtualMemory failed
            #
            # c: ... \lib\site-packages\psutil\_pswindows.py:806: RuntimeError
            cmdline = None

        if not cmdline:  # pragma: no cover
            try:
                cmdline = proc.as_dict()
            except psutil.NoSuchProcess:
                cmdline = f"<could not be retrived; dead process: {proc}>"
            except (psutil.AccessDenied, OSError):  # pragma: no cover
                cmdline = weakref.proxy(proc)
        proc._cmdline = cmdline
    return proc._cmdline
    # pylint: enable=protected-access


def _terminate_process_list(
    process_list: List[psutil.Process], kill: bool = False, slow_stop: bool = False
) -> None:
    log.info(
        "Terminating process list:\n%s",
        pprint.pformat([_get_cmdline(proc) for proc in process_list]),
    )
    for process in process_list[:]:  # Iterate over copy of the list
        if not psutil.pid_exists(process.pid):
            process_list.remove(process)
            continue
        try:
            if not kill and process.status() == psutil.STATUS_ZOMBIE:  # pragma: no cover
                # Zombie processes will exit once child processes also exit
                continue
            if kill:
                log.info("Killing process(%s): %s", process.pid, _get_cmdline(process))
                process.kill()
            else:
                log.info("Terminating process(%s): %s", process.pid, _get_cmdline(process))
                try:
                    if slow_stop:
                        # Allow coverage data to be written down to disk
                        process.send_signal(signal.SIGTERM)
                        try:
                            process.wait(2)
                        except psutil.TimeoutExpired:  # pragma: no cover
                            if psutil.pid_exists(process.pid):
                                continue
                    else:
                        process.terminate()
                except OSError as exc:  # pragma: no cover
                    if exc.errno not in (errno.ESRCH, errno.EACCES):
                        raise
            if not psutil.pid_exists(process.pid):
                process_list.remove(process)
        except psutil.NoSuchProcess:
            process_list.remove(process)


def terminate_process_list(
    process_list: List[psutil.Process], kill: bool = False, slow_stop: bool = False
) -> None:
    """
    Terminate a list of processes.

    Arguments:
        process_list:
            An iterable of :py:class:`psutil.Process` instances to terminate

    Keyword Arguments:
        kill:
            Kill the process instead of terminating it.
        slow_stop:
            First try to terminate each process in the list, and if termination was not successful, kill it.

    Returns:
        Nothing.
    """

    def on_process_terminated(proc: psutil.Process) -> None:
        log.info(
            "Process %s terminated with exit code: %s",
            getattr(proc, "_cmdline", proc),
            proc.returncode,
        )

    # Try to terminate processes with the provided kill and slow_stop parameters
    log.info("Terminating process list. 1st step. kill: %s, slow stop: %s", kill, slow_stop)

    # Remove duplicates from the process list
    seen_pids = []
    start_count = len(process_list)
    for proc in process_list[:]:
        if proc.pid in seen_pids:
            process_list.remove(proc)
        seen_pids.append(proc.pid)
    end_count = len(process_list)
    if end_count < start_count:
        log.debug("Removed %d duplicates from the initial process list", start_count - end_count)

    _terminate_process_list(process_list, kill=kill, slow_stop=slow_stop)
    psutil.wait_procs(process_list, timeout=5, callback=on_process_terminated)

    if process_list:
        # If there's still processes to be terminated, retry and kill them if slow_stop is False
        log.info(
            "Terminating process list. 2nd step. kill: %s, slow stop: %s",
            slow_stop is False,
            slow_stop,
        )
        _terminate_process_list(process_list, kill=slow_stop is False, slow_stop=slow_stop)
        psutil.wait_procs(process_list, timeout=5, callback=on_process_terminated)

    if process_list:
        # If there's still processes to be terminated, just kill them, no slow stopping now
        log.info("Terminating process list. 3rd step. kill: True, slow stop: False")
        _terminate_process_list(process_list, kill=True, slow_stop=False)
        psutil.wait_procs(process_list, timeout=5, callback=on_process_terminated)

    if process_list:
        # In there's still processes to be terminated, log a warning about it
        log.warning("Some processes failed to properly terminate: %s", process_list)


def terminate_process(
    pid: Optional[int] = None,
    process: Optional[psutil.Process] = None,
    children: Optional[List[psutil.Process]] = None,
    kill_children: Optional[bool] = None,
    slow_stop: bool = False,
) -> None:
    """
    Try to terminate/kill the started process.

    Keyword Arguments:
        pid:
            The PID of the process
        process:
            An instance of :py:class:`psutil.Process`
        children:
            An iterable of :py:class:`psutil.Process` instances, children to the process being terminated
        kill_children:
            Also try to terminate/kill child processes
        slow_stop:
            First try to terminate each process in the list, and if termination was not successful, kill it.
    """
    children = children or []
    process_list: List[psutil.Process] = []

    if kill_children is None:
        # Always kill children if kill the parent process and kill_children was not set
        kill_children = True if slow_stop is False else kill_children

    if pid and not process:
        try:
            process = psutil.Process(pid)
            process_list.append(process)
        except psutil.NoSuchProcess:
            # Process is already gone
            process = None

    if kill_children:
        if process:
            children.extend(collect_child_processes(process.pid))
        if children:
            process_list.extend(children)

    if process_list:
        if process:
            log.info("Stopping process %s and respective children: %s", process, children)
        else:
            log.info("Terminating process list: %s", process_list)
        terminate_process_list(process_list, kill=slow_stop is False, slow_stop=slow_stop)



# Copyright 2022-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import textwrap

import pytest

import pytestshellutils.exceptions as exceptions
from pytestshellutils.utils.processes import ProcessResult


def test_process_failed_message() -> None:
    message = "The message"
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message)
    assert str(exc.value) == message


def test_process_failed_cmdline() -> None:
    message = "The message"
    cmdline = ["python", "--version"]
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Command Line: {!r}
         Returncode: 0
    """.format(
            message, cmdline
        )
    )
    pres = ProcessResult(returncode=0, cmdline=cmdline, stdout="", stderr="")
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected


def test_process_failed_returncode() -> None:
    message = "The message"
    returncode = 1
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Returncode: {}
    """.format(
            message, returncode
        )
    )
    pres = ProcessResult(returncode=returncode, stdout="", stderr="")
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected


def test_process_failed_stdout() -> None:
    message = "The message"
    stdout = "This is the STDOUT"
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Returncode: 0
         Process Output:
           >>>>> STDOUT >>>>>
        {}
           <<<<< STDOUT <<<<<
    """.format(
            message, stdout
        )
    )
    pres = ProcessResult(returncode=0, stdout=stdout, stderr="")
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected


def test_process_failed_stderr() -> None:
    message = "The message"
    stderr = "This is the STDERR"
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Returncode: 0
         Process Output:
           >>>>> STDERR >>>>>
        {}
           <<<<< STDERR <<<<<
    """.format(
            message, stderr
        )
    )
    pres = ProcessResult(returncode=0, stdout="", stderr=stderr)
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected


def test_process_failed_stdout_and_stderr() -> None:
    message = "The message"
    stdout = "This is the STDOUT"
    stderr = "This is the STDERR"
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Returncode: 0
         Process Output:
           >>>>> STDOUT >>>>>
        {}
           <<<<< STDOUT <<<<<
           >>>>> STDERR >>>>>
        {}
           <<<<< STDERR <<<<<
    """.format(
            message, stdout, stderr
        )
    )
    pres = ProcessResult(returncode=0, stdout=stdout, stderr=stderr)
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected


def test_process_failed_cmdline_stdout_and_stderr() -> None:
    message = "The message"
    stdout = "This is the STDOUT"
    stderr = "This is the STDERR"
    cmdline = ["python", "--version"]
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Command Line: {!r}
         Returncode: 0
         Process Output:
           >>>>> STDOUT >>>>>
        {}
           <<<<< STDOUT <<<<<
           >>>>> STDERR >>>>>
        {}
           <<<<< STDERR <<<<<
    """.format(
            message, cmdline, stdout, stderr
        )
    )
    pres = ProcessResult(returncode=0, stdout=stdout, stderr=stderr, cmdline=cmdline)
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected


def test_process_failed_cmdline_stdout_stderr_and_returncode() -> None:
    message = "The message"
    stdout = "This is the STDOUT"
    stderr = "This is the STDERR"
    cmdline = ["python", "--version"]
    returncode = 1
    expected = textwrap.dedent(
        """\
        {}
        ProcessResult
         Command Line: {!r}
         Returncode: {}
         Process Output:
           >>>>> STDOUT >>>>>
        {}
           <<<<< STDOUT <<<<<
           >>>>> STDERR >>>>>
        {}
           <<<<< STDERR <<<<<
    """.format(
            message, cmdline, returncode, stdout, stderr
        )
    )
    pres = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline)
    with pytest.raises(exceptions.FactoryFailure) as exc:
        raise exceptions.FactoryFailure(message, process_result=pres)
    output = str(exc.value)
    assert output == expected



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Pytest Shell Utilities related exceptions.
"""
from typing import Optional

from pytestshellutils.utils.processes import ProcessResult


class ShellUtilsException(Exception):
    """
    Base pytest shell utilities exception.
    """


class CallbackException(ShellUtilsException):
    """
    Exception raised during a before/after start/stop daemon callback.
    """


class ProcessFailed(ShellUtilsException):
    """
    Exception raised when a sub-process fails.

    Arguments:
        message:
            The exception message

    Keyword Arguments:
        process_result:
            The ``ProcessResult`` instance when the exception occurred
    """

    def __init__(self, message: str, process_result: Optional[ProcessResult] = None) -> None:
        super().__init__()
        self.message = message
        self.process_result = process_result

    def __str__(self) -> str:
        """
        Return a printable representation of the exception.
        """
        message = self.message
        if self.process_result:
            if not message.endswith("\n"):
                message += "\n"
            message += str(self.process_result)
        return message


class FactoryFailure(ProcessFailed):
    """
    Exception raised when a sub-process fails on one of the factories.
    """


class FactoryNotStarted(FactoryFailure):
    """
    Exception raised when a factory failed to start.

    Please look at :py:class:`~pytestshellutils.exceptions.FactoryFailure` for the supported keyword
    arguments documentation.
    """


class FactoryNotRunning(FactoryFailure):
    """
    Exception raised when trying to use a factory's `.stopped` context manager and the factory is not running.

    Please look at :py:class:`~pytestshellutils.exceptions.FactoryFailure` for the supported keyword
    arguments documentation.
    """


class ProcessNotStarted(FactoryFailure):
    """
    Exception raised when a process failed to start.

    Please look at :py:class:`~pytestshellutils.exceptions.FactoryFailure` for the supported keywords.
    arguments documentation.
    """


class FactoryTimeout(FactoryNotStarted):
    """
    Exception raised when a process timed-out.

    Please look at :py:class:`~pytestshellutils.exceptions.FactoryFailure` for the supported keywords.
    arguments documentation.
    """



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import os
import pathlib
import sys
from typing import Any
from typing import cast

import pytest
from pytest_subtests import SubTests

from pytestshellutils.customtypes import EnvironDict
from pytestshellutils.exceptions import FactoryTimeout
from pytestshellutils.shell import ScriptSubprocess
from tests.conftest import Tempfiles


@pytest.mark.parametrize("exitcode", [0, 1, 3, 9, 40, 120])
def test_exitcode(exitcode: int, tempfiles: Tempfiles) -> None:
    shell = ScriptSubprocess(script_name=sys.executable)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.125)
        exit({})
        """.format(
            exitcode
        )
    )
    result = shell.run(script)
    assert result.returncode == exitcode


def test_timeout_defined_on_class_instantiation(tempfiles: Tempfiles) -> None:
    shell = ScriptSubprocess(script_name=sys.executable, timeout=0.5)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(1)
        exit(0)
        """
    )
    with pytest.raises(FactoryTimeout):
        shell.run(script)


def test_timeout_defined_run(tempfiles: Tempfiles) -> None:
    shell = ScriptSubprocess(script_name=sys.executable)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.5)
        exit(0)
        """
    )
    result = shell.run(script)
    assert result.returncode == 0

    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.5)
        exit(0)
        """
    )
    with pytest.raises(FactoryTimeout):
        shell.run(script, _timeout=0.1)


@pytest.mark.parametrize(
    "input_str,expected_object",
    [
        # Good JSON
        ('{"a": "a", "1": 1}', {"a": "a", "1": 1}),
        # Bad JSON
        ("{'a': 'a', '1': 1}", None),
    ],
)
def test_json_output(input_str: str, expected_object: Any, tempfiles: Tempfiles) -> None:
    shell = ScriptSubprocess(script_name=sys.executable)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import sys
        sys.stdout.write('''{}''')
        exit(0)
        """.format(
            input_str
        )
    )
    result = shell.run(script)
    assert result.returncode == 0
    if expected_object:
        assert result.data == expected_object
    assert result.stdout == input_str


def test_stderr_output(tempfiles: Tempfiles) -> None:
    input_str = "Thou shalt not exit cleanly"
    shell = ScriptSubprocess(script_name=sys.executable)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        exit("{}")
        """.format(
            input_str
        )
    )
    result = shell.run(script)
    assert result.returncode == 1
    assert result.stderr == input_str + "\n"


def test_unicode_output(tempfiles: Tempfiles) -> None:
    shell = ScriptSubprocess(script_name=sys.executable)
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8
        from __future__ import print_function
        import sys
        sys.stdout.write(u'STDOUT F\xe1tima')
        sys.stdout.flush()
        sys.stderr.write(u'STDERR F\xe1tima')
        sys.stderr.flush()
        exit(0)
        """
    )
    result = shell.run(script)
    assert result.returncode == 0, str(result)
    assert result.stdout == "STDOUT Fátima"
    assert result.stderr == "STDERR Fátima"


def test_process_failed_to_start(tempfiles: Tempfiles) -> None:
    shell = ScriptSubprocess(script_name=sys.executable)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        1/0
        """
    )
    result = shell.run(script)
    assert result.returncode == 1
    assert "ZeroDivisionError: division by zero" in result.stderr


def test_environ(tempfiles: Tempfiles) -> None:
    environ = cast(EnvironDict, os.environ.copy())
    environ["FOO"] = "foo"
    shell = ScriptSubprocess(script_name=sys.executable, environ=environ)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import os
        import json
        print(json.dumps(dict(**os.environ)), flush=True)
        exit(0)
        """
    )
    result = shell.run(script)
    assert result.returncode == 0
    assert result.data
    assert "FOO" in result.data
    assert result.data["FOO"] == "foo"


def test_env_in_run_call(tempfiles: Tempfiles) -> None:
    env = cast(EnvironDict, {"FOO": "bar"})
    environ = cast(EnvironDict, os.environ.copy())
    environ["FOO"] = "foo"
    shell = ScriptSubprocess(script_name=sys.executable, environ=environ)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import os
        import json
        print(json.dumps(dict(**os.environ)), flush=True)
        exit(0)
        """
    )
    result = shell.run(script, env=env)
    assert result.returncode == 0
    assert result.data
    assert "FOO" in result.data
    assert result.data["FOO"] == env["FOO"]


def test_not_started() -> None:
    shell = ScriptSubprocess(script_name=sys.executable)
    assert shell.is_running() is False
    assert not shell.pid
    assert shell.terminate() is None


def test_display_name(tempfiles: Tempfiles) -> None:
    python_binary_name = pathlib.Path(sys.executable).name
    display_name = "ScriptSubprocess({})".format(python_binary_name)
    shell = ScriptSubprocess(script_name=sys.executable)
    assert shell.get_display_name() == display_name
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.125)
        exit(0)
        """
    )
    result = shell.run(script)
    assert result.returncode == 0
    assert shell.get_display_name() == display_name


def test_get_script_path(subtests: SubTests) -> None:
    python_binary_name = pathlib.Path(sys.executable).name
    with subtests.test(script_name=sys.executable):
        shell = ScriptSubprocess(script_name=sys.executable)
        assert shell.get_script_path() == sys.executable
    with subtests.test(script_name=python_binary_name):
        shell = ScriptSubprocess(script_name=python_binary_name)
        assert shell.get_script_path() == sys.executable
    python_binary_name = "{}3.100".format(python_binary_name)
    with subtests.test(script_name=python_binary_name):
        shell = ScriptSubprocess(script_name=python_binary_name)
        with pytest.raises(FileNotFoundError) as exc:
            shell.get_script_path()
        assert "The CLI script {!r} does not exist".format(python_binary_name) in str(exc)



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Test ``pytestshellutils.utils.processes.ProcessResult``.
"""
import json as jsonlib
import logging
import textwrap

import pytest
from pytest_subtests import SubTests

from pytestshellutils.utils.processes import ProcessResult

log = logging.getLogger(__name__)


@pytest.mark.parametrize("returncode", [None, 1.0, -1.0, "0"])
def test_non_int_returncode_raises_exception(returncode: int) -> None:
    with pytest.raises(ValueError):
        ProcessResult(returncode=returncode, stdout="", stderr="")


def test_attributes(subtests: SubTests) -> None:
    returncode = 0
    stdout = "STDOUT"
    stderr = "STDERR"
    cmdline = None
    data = None

    with subtests.test(returncode=returncode, stdout=stdout, stderr=stderr):
        ret = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr)
        assert ret.returncode == returncode
        assert ret.stdout == stdout
        assert ret.stderr == stderr
        assert ret.data == data
        assert ret.cmdline == cmdline

    cmdline = ["1", "2", "3"]
    with subtests.test(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline):
        ret = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline)
        assert ret.returncode == returncode
        assert ret.stdout == stdout
        assert ret.stderr == stderr
        assert ret.data == data
        assert ret.cmdline == cmdline

    data = {"ret": {"a": 1}}
    stdout = jsonlib.dumps(data)
    with subtests.test(
        returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline, data=data
    ):
        ret = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline)
        assert ret.returncode == returncode
        assert ret.stdout == stdout
        assert ret.stderr == stderr
        assert ret.data == data
        assert ret.cmdline == cmdline

    data_key = "ret"
    with subtests.test(
        returncode=returncode,
        stdout=stdout,
        stderr=stderr,
        cmdline=cmdline,
        data=data,
        data_key=data_key,
    ):
        ret = ProcessResult(
            returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline, data_key="ret"
        )
        assert ret.returncode == returncode
        assert ret.stdout == stdout
        assert ret.stderr == stderr
        assert ret.data == data[data_key]
        assert ret.cmdline == cmdline


def test_str_formatting(subtests: SubTests) -> None:
    returncode = 0
    stdout = "STDOUT"
    stderr = "STDERR"
    cmdline = None
    data = {"ret": {"a": 1}}
    data_key = None

    with subtests.test(returncode=returncode, stdout=stdout, stderr=stderr):
        ret = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr)
        expected = textwrap.dedent(
            """\
            ProcessResult
             Returncode: {}
             Process Output:
               >>>>> STDOUT >>>>>
            {}
               <<<<< STDOUT <<<<<
               >>>>> STDERR >>>>>
            {}
               <<<<< STDERR <<<<<
        """.format(
                returncode, stdout, stderr
            )
        )
        assert str(ret) == expected

    cmdline = ["1", "2", "3"]
    with subtests.test(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline):
        ret = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline)
        expected = textwrap.dedent(
            """\
            ProcessResult
             Command Line: {!r}
             Returncode: {}
             Process Output:
               >>>>> STDOUT >>>>>
            {}
               <<<<< STDOUT <<<<<
               >>>>> STDERR >>>>>
            {}
               <<<<< STDERR <<<<<
        """.format(
                cmdline,
                returncode,
                stdout,
                stderr,
            )
        )
        assert str(ret) == expected

    stdout = jsonlib.dumps(data)
    with subtests.test(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline):
        ret = ProcessResult(returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline)
        expected = (
            textwrap.dedent(
                f"""\
            ProcessResult
             Command Line: {cmdline!r}
             Returncode: {returncode}
             Process Output:
               >>>>> STDOUT >>>>>
            {stdout}
               <<<<< STDOUT <<<<<
               >>>>> STDERR >>>>>
            {stderr}
               <<<<< STDERR <<<<<
             Parsed JSON Data:
        """
            )
            + ProcessResult._to_printable_data(data)
            + "\n"
        )
        assert str(ret) == expected

    data_key = "ret"
    with subtests.test(
        returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline, data_key=data_key
    ):
        ret = ProcessResult(
            returncode=returncode, stdout=stdout, stderr=stderr, cmdline=cmdline, data_key=data_key
        )
        expected = (
            textwrap.dedent(
                f"""\
            ProcessResult
             Command Line: {cmdline!r}
             Returncode: {returncode}
             Process Output:
               >>>>> STDOUT >>>>>
            {stdout}
               <<<<< STDOUT <<<<<
               >>>>> STDERR >>>>>
            {stderr}
               <<<<< STDERR <<<<<
             Parsed JSON Data:
        """
            )
            + ProcessResult._to_printable_data(data[data_key])
            + "\n"
        )
        assert str(ret) == expected



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
from typing import Any
from typing import Dict

import pytest

from pytestshellutils.customtypes import Callback


def func(*args: Any, **kwargs: Any) -> Dict[str, Any]:
    return {"args": args, "kwargs": kwargs}


@pytest.fixture
def callback() -> Callback:
    return Callback(func=func, args=("a1", "a2"), kwargs={"keyword_argument": True})


def test___str__(callback: Callback) -> None:
    assert str(callback) == "func('a1', 'a2', keyword_argument=True)"


def test___call__(callback: Callback) -> None:
    assert callback() == {"args": ("a1", "a2"), "kwargs": {"keyword_argument": True}}


def test___call__extra_args(callback: Callback) -> None:
    assert callback("bar") == {"args": ("bar", "a1", "a2"), "kwargs": {"keyword_argument": True}}


def test___call__extra_kwargs(callback: Callback) -> None:
    assert callback(bar=2) == {"args": ("a1", "a2"), "kwargs": {"keyword_argument": True, "bar": 2}}


def test___call__override_kwarg(callback: Callback) -> None:
    assert callback(keyword_argument=False) == {
        "args": ("a1", "a2"),
        "kwargs": {"keyword_argument": False},
    }



# Copyright 2022-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import os
import pathlib
import sys
import tempfile
from typing import Any
from typing import cast

import pytest

from pytestshellutils.customtypes import EnvironDict
from pytestshellutils.exceptions import FactoryTimeout
from pytestshellutils.shell import Subprocess
from tests.conftest import Tempfiles


@pytest.mark.parametrize("exitcode", [0, 1, 3, 9, 40, 120])
def test_exitcode(exitcode: int, tempfiles: Tempfiles) -> None:
    shell = Subprocess()
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.125)
        exit({})
        """.format(
            exitcode
        )
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == exitcode


def test_timeout_defined_on_class_instantiation(tempfiles: Tempfiles) -> None:
    shell = Subprocess(timeout=0.5)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(1)
        exit(0)
        """
    )
    with pytest.raises(FactoryTimeout):
        shell.run(sys.executable, script)


def test_timeout_defined_run(tempfiles: Tempfiles) -> None:
    shell = Subprocess()
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.5)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 0

    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.5)
        exit(0)
        """
    )
    with pytest.raises(FactoryTimeout):
        shell.run(sys.executable, script, _timeout=0.1)


@pytest.mark.parametrize(
    "input_str,expected_object",
    [
        # Good JSON
        ('{"a": "a", "1": 1}', {"a": "a", "1": 1}),
        # Bad JSON
        ("{'a': 'a', '1': 1}", None),
    ],
)
def test_json_output(input_str: str, expected_object: Any, tempfiles: Tempfiles) -> None:
    shell = Subprocess()
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import sys
        sys.stdout.write('''{}''')
        exit(0)
        """.format(
            input_str
        )
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 0
    if expected_object:
        assert result.data == expected_object
    assert result.stdout == input_str


def test_stderr_output(tempfiles: Tempfiles) -> None:
    input_str = "Thou shalt not exit cleanly"
    shell = Subprocess()
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        exit("{}")
        """.format(
            input_str
        )
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 1
    assert result.stderr == input_str + "\n"


def test_unicode_output(tempfiles: Tempfiles) -> None:
    shell = Subprocess()
    script = tempfiles.makepyfile(
        r"""
        # coding=utf-8
        from __future__ import print_function
        import sys
        sys.stdout.write(u'STDOUT F\xe1tima')
        sys.stdout.flush()
        sys.stderr.write(u'STDERR F\xe1tima')
        sys.stderr.flush()
        exit(0)
        """
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 0, str(result)
    assert result.stdout == "STDOUT Fátima"
    assert result.stderr == "STDERR Fátima"


def test_process_failed_to_start(tempfiles: Tempfiles) -> None:
    shell = Subprocess()
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        1/0
        """
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 1
    assert "ZeroDivisionError: division by zero" in result.stderr


def test_environ(tempfiles: Tempfiles) -> None:
    environ = cast(EnvironDict, os.environ.copy())
    environ["FOO"] = "foo"
    shell = Subprocess(environ=environ)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import os
        import json
        print(json.dumps(dict(**os.environ)), flush=True)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 0
    assert result.data
    assert "FOO" in result.data
    assert result.data["FOO"] == "foo"


def test_env_in_run_call(tempfiles: Tempfiles) -> None:
    env = cast(EnvironDict, {"FOO": "bar"})
    environ = cast(EnvironDict, os.environ.copy())
    environ["FOO"] = "foo"
    shell = Subprocess(environ=environ)
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import os
        import json
        print(json.dumps(dict(**os.environ)), flush=True)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script, env=env)
    assert result.returncode == 0
    assert result.data
    assert "FOO" in result.data
    assert result.data["FOO"] == env["FOO"]


def test_not_started() -> None:
    shell = Subprocess()
    assert shell.is_running() is False
    assert not shell.pid
    assert shell.terminate() is None


def test_display_name(tempfiles: Tempfiles) -> None:
    shell = Subprocess()
    assert shell.get_display_name() == "Subprocess()"
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import time
        time.sleep(0.125)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 0
    assert shell.get_display_name() == "Subprocess([{!r}, {!r}])".format(sys.executable, script)


def test_run_cwd(tmp_path: str, tempfiles: Tempfiles, shell: Subprocess) -> None:
    shell = Subprocess(cwd=tmp_path)
    system_tempdir = str(pathlib.Path(tempfile.gettempdir()).resolve())
    assert str(shell.cwd.resolve()) != system_tempdir
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import pathlib
        print(str(pathlib.Path.cwd().resolve()), flush=True)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script, cwd=system_tempdir)
    assert result.returncode == 0
    assert result.stdout.strip() == system_tempdir


def test_run_shell() -> None:
    shell = Subprocess()
    with pytest.raises(FileNotFoundError):
        shell.run("exit", "0")
    with pytest.raises(FileNotFoundError):
        shell.run("exit", "0", shell=False)
    result = shell.run("exit", "0", shell=True)
    assert result.returncode == 0



# Copyright 2022-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib
import sys
import tempfile

import pytest

from pytestshellutils.shell import Subprocess
from tests.conftest import Tempfiles


def test_run_call(tempfiles: Tempfiles, shell: Subprocess) -> None:
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import os
        import json
        print(json.dumps(dict(**os.environ)), flush=True)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script)
    assert result.returncode == 0


def test_run_cwd(tempfiles: Tempfiles, shell: Subprocess) -> None:
    system_tempdir = str(pathlib.Path(tempfile.gettempdir()).resolve())
    assert str(shell.cwd.resolve()) != system_tempdir
    script = tempfiles.makepyfile(
        """
        # coding=utf-8
        import pathlib
        print(str(pathlib.Path.cwd().resolve()), flush=True)
        exit(0)
        """
    )
    result = shell.run(sys.executable, script, cwd=system_tempdir)
    assert result.returncode == 0
    assert result.stdout.strip() == system_tempdir


def test_run_shell(shell: Subprocess) -> None:
    with pytest.raises(FileNotFoundError):
        shell.run("exit", "0")
    with pytest.raises(FileNotFoundError):
        shell.run("exit", "0", shell=False)
    result = shell.run("exit", "0", shell=True)
    assert result.returncode == 0



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
"""
Pytest shell utilities plugin.
"""
import pytest

from pytestshellutils.shell import Subprocess


@pytest.fixture(scope="session")
def shell() -> Subprocess:
    """
    Shell fixture.

    Example:
        .. code-block:: python

           def test_assert_good_exitcode(shell):

               ret = shell.run("exit", "0")
               assert ret.returncode == 0


           def test_assert_bad_exitcode(shell):

               ret = shell.run("exit", "1")
               assert ret.returncode == 1
    """
    return Subprocess()



# Copyright 2022-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import time
from unittest import mock

import pytestshellutils.utils.time


def test_sleep() -> None:
    start = time.time()
    with mock.patch("time.sleep", return_value=None):
        time.sleep(1)
        pytestshellutils.utils.time.sleep(0.1)
    end = time.time()
    duration = end - start
    assert duration >= 0.1  # We did sleep 0.1 second
    assert duration < 0.5  # But the patched time.sleep was mocked



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
from pytestshellutils.utils import format_callback_to_string


def test_format_from_string() -> None:
    func = "the_function"
    args = ("one", "two")
    kwargs = {"three": 3}
    assert (
        format_callback_to_string(func, args=args, kwargs=kwargs)
        == "the_function('one', 'two', three=3)"
    )


def test_format_just_args() -> None:
    func = "the_function"
    args = ("one", "two")
    assert format_callback_to_string(func, args=args) == "the_function('one', 'two')"


def test_format_just_kwargs() -> None:
    func = "the_function"
    kwargs = {"three": 3}
    assert format_callback_to_string(func, kwargs=kwargs) == "the_function(three=3)"


def test_format_no_args_nor_kwargs() -> None:
    func = "the_function"
    assert format_callback_to_string(func) == "the_function()"


def test_format_from_function() -> None:
    func = format_callback_to_string
    args = ("one", "two")
    kwargs = {"three": 3}
    assert (
        format_callback_to_string(func, args, kwargs)
        == "format_callback_to_string('one', 'two', three=3)"
    )



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=import-error,protected-access,line-too-long
import datetime
import gzip
import json
import os
import pathlib
import shutil
import sys
import tarfile
import tempfile

import nox
from nox.command import CommandFailed


COVERAGE_VERSION_REQUIREMENT = "coverage==7.2.7"
PYTEST_VERSION_REQUIREMENT = os.environ.get("PYTEST_VERSION_REQUIREMENT") or None
IS_WINDOWS = sys.platform.lower().startswith("win")
IS_DARWIN = sys.platform.lower().startswith("darwin")

if IS_WINDOWS:
    COVERAGE_FAIL_UNDER_PERCENT = 87
elif IS_DARWIN:
    COVERAGE_FAIL_UNDER_PERCENT = 87
else:
    COVERAGE_FAIL_UNDER_PERCENT = 87

# Be verbose when running under a CI context
PIP_INSTALL_SILENT = (os.environ.get("CI") or os.environ.get("GITHUB_ACTIONS")) is None
CI_RUN = PIP_INSTALL_SILENT is False
SKIP_REQUIREMENTS_INSTALL = "SKIP_REQUIREMENTS_INSTALL" in os.environ
EXTRA_REQUIREMENTS_INSTALL = os.environ.get("EXTRA_REQUIREMENTS_INSTALL")

# Paths
REPO_ROOT = pathlib.Path(__file__).resolve().parent
# Change current directory to REPO_ROOT
os.chdir(str(REPO_ROOT))

SITECUSTOMIZE_DIR = str(REPO_ROOT / "tests" / "support" / "coverage")
ARTIFACTS_DIR = REPO_ROOT / "artifacts"
# Make sure the artifacts directory exists
ARTIFACTS_DIR.mkdir(parents=True, exist_ok=True)
RUNTESTS_LOGFILE = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "runtests-{}.log".format(
    datetime.datetime.now().strftime("%Y%m%d%H%M%S.%f")
)
COVERAGE_REPORT_DB = REPO_ROOT / ".coverage"
COVERAGE_REPORT_PROJECT = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "coverage-project.xml"
COVERAGE_REPORT_TESTS = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "coverage-tests.xml"
JUNIT_REPORT = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "junit-report.xml"

# Nox options
#  Reuse existing virtualenvs
nox.options.reuse_existing_virtualenvs = True
#  Don't fail on missing interpreters
nox.options.error_on_missing_interpreters = False


def pytest_version(session):
    try:
        return session._runner._pytest_version_info
    except AttributeError:
        session_pytest_version = session_run_always(
            session,
            "python",
            "-c",
            'import sys, pkg_resources; sys.stdout.write("{}".format(pkg_resources.get_distribution("pytest").version))',
            silent=True,
            log=False,
        )
        session._runner._pytest_version_info = tuple(
            int(part) for part in session_pytest_version.split(".") if part.isdigit()
        )
    return session._runner._pytest_version_info


def python_version(session):
    try:
        return session._runner._python_version_info
    except AttributeError:
        session_python_version = session_run_always(
            session,
            "python",
            "-c",
            'import sys; sys.stdout.write("{}.{}".format(*sys.version_info));',
            silent=True,
            log=False,
        )
        session._runner._python_version_info = tuple(
            int(part) for part in session_python_version.split(".") if part.isdigit()
        )
    return session._runner._python_version_info


def session_run_always(session, *command, **kwargs):
    try:
        # Guess we weren't the only ones wanting this
        # https://github.com/theacodes/nox/pull/331
        return session.run_always(*command, **kwargs)
    except AttributeError:
        old_install_only_value = session._runner.global_config.install_only
        try:
            # Force install only to be false for the following chunk of code
            # For additional information as to why see:
            #   https://github.com/theacodes/nox/pull/181
            session._runner.global_config.install_only = False
            return session.run(*command, **kwargs)
        finally:
            session._runner.global_config.install_only = old_install_only_value


@nox.session(python=("3", "3.7", "3.8", "3.9", "3.10", "3.11"))
def tests(session):
    """
    Run tests.
    """
    env = {}
    install_arguments = ["--progress-bar=off"]
    if SKIP_REQUIREMENTS_INSTALL is False:
        env["COVERAGE_PLUGINS_LIST"] = "coverage_conditional_plugin"
        coverage_requirements = [COVERAGE_VERSION_REQUIREMENT, "coverage-conditional-plugin"]
        session.install(*install_arguments, "wheel", silent=PIP_INSTALL_SILENT)
        session.install(
            *install_arguments,
            *coverage_requirements,
            silent=PIP_INSTALL_SILENT,
        )

        pytest_version_requirement = PYTEST_VERSION_REQUIREMENT
        if pytest_version_requirement:
            if not pytest_version_requirement.startswith("pytest"):
                if (
                    not pytest_version_requirement.startswith(("==", "~="))
                    and pytest_version_requirement.endswith(".0")
                    and not pytest_version_requirement.startswith("~=")
                ):
                    pytest_version_requirement = f"~={pytest_version_requirement}"
                pytest_version_requirement = f"pytest{pytest_version_requirement}"
            session.install(pytest_version_requirement, silent=PIP_INSTALL_SILENT)
        session.install(*install_arguments, "-e", ".[tests]", silent=PIP_INSTALL_SILENT)

        if EXTRA_REQUIREMENTS_INSTALL:
            session.log(
                "Installing the following extra requirements because the EXTRA_REQUIREMENTS_INSTALL "
                "environment variable was set: EXTRA_REQUIREMENTS_INSTALL='%s'",
                EXTRA_REQUIREMENTS_INSTALL,
            )
            install_command = [req.strip() for req in EXTRA_REQUIREMENTS_INSTALL.split()]
            session.install(*install_arguments, *install_command, silent=PIP_INSTALL_SILENT)

    session.run("coverage", "erase")

    python_path_env_var = os.environ.get("PYTHONPATH") or None
    if python_path_env_var is None:
        python_path_env_var = SITECUSTOMIZE_DIR
    else:
        python_path_entries = python_path_env_var.split(os.pathsep)
        if SITECUSTOMIZE_DIR in python_path_entries:
            python_path_entries.remove(SITECUSTOMIZE_DIR)
        python_path_entries.insert(0, SITECUSTOMIZE_DIR)
        python_path_env_var = os.pathsep.join(python_path_entries)

    env.update(
        {
            # The updated python path so that sitecustomize is importable
            "PYTHONPATH": python_path_env_var,
            # The full path to the .coverage data file. Makes sure we always write
            # them to the same directory
            "COVERAGE_FILE": str(COVERAGE_REPORT_DB),
            # Instruct sub processes to also run under coverage
            "COVERAGE_PROCESS_START": str(REPO_ROOT / ".coveragerc"),
        }
    )

    args = [
        "--rootdir",
        str(REPO_ROOT),
        "--log-file={}".format(RUNTESTS_LOGFILE),
        "--log-file-level=debug",
        "--show-capture=no",
        "--junitxml={}".format(JUNIT_REPORT),
        "--showlocals",
        "--strict-markers",
        "-ra",
        "-s",
    ]
    if pytest_version(session) > (6, 2):
        args.append("--lsof")
    if session._runner.global_config.forcecolor:
        args.append("--color=yes")
    if not session.posargs:
        args.append("tests/")
    else:
        for arg in session.posargs:
            if arg.startswith("--color") and session._runner.global_config.forcecolor:
                args.remove("--color=yes")
            args.append(arg)

    try:
        session.run("python", "-bb", "-m", "coverage", "run", "-m", "pytest", *args, env=env)
    finally:
        # Always combine and generate the XML coverage report
        try:
            session.run("coverage", "combine")
        except CommandFailed:
            # Sometimes some of the coverage files are corrupt which would
            # trigger a CommandFailed exception
            pass
        # Generate report for project code coverage
        session.run(
            "coverage",
            "xml",
            "-o",
            str(COVERAGE_REPORT_PROJECT),
            "--omit=tests/*",
            "--include=src/pytestshellutils/*",
        )
        # Generate report for tests code coverage
        session.run(
            "coverage",
            "xml",
            "-o",
            str(COVERAGE_REPORT_TESTS),
            "--omit=src/pytestshellutils/*",
            "--include=tests/*",
        )
        try:
            cmdline = [
                "coverage",
                "report",
                "--show-missing",
                "--include=src/pytestshellutils/*,tests/*",
            ]
            if pytest_version(session) >= (6, 2):
                cmdline.append("--fail-under={}".format(COVERAGE_FAIL_UNDER_PERCENT))
            session.run(*cmdline)
        finally:
            if COVERAGE_REPORT_DB.exists():
                shutil.copyfile(str(COVERAGE_REPORT_DB), str(ARTIFACTS_DIR / ".coverage"))


def _lint(session, rcfile, flags, paths):
    session.install("--progress-bar=off", "-e", ".[lint]", silent=PIP_INSTALL_SILENT)
    session.run("pylint", "--version")
    pylint_report_path = os.environ.get("PYLINT_REPORT")

    cmd_args = ["pylint", "--rcfile={}".format(rcfile)] + list(flags) + list(paths)

    stdout = tempfile.TemporaryFile(mode="w+b")
    try:
        session.run(*cmd_args, stdout=stdout)
    finally:
        stdout.seek(0)
        contents = stdout.read()
        if contents:
            contents = contents.decode("utf-8")
            sys.stdout.write(contents)
            sys.stdout.flush()
            if pylint_report_path:
                # Write report
                with open(pylint_report_path, "w", encoding="utf-8") as wfh:
                    wfh.write(contents)
                session.log("Report file written to %r", pylint_report_path)
        stdout.close()


@nox.session(python="3")
def lint(session):
    """
    Run PyLint against Salt and it's test suite. Set PYLINT_REPORT to a path to capture output.
    """
    session.notify("lint-code-{}".format(session.python))
    session.notify("lint-tests-{}".format(session.python))


@nox.session(python="3", name="lint-code")
def lint_code(session):
    """
    Run PyLint against the code. Set PYLINT_REPORT to a path to capture output.
    """
    flags = ["--disable=I"]
    if session.posargs:
        paths = session.posargs
    else:
        paths = ["setup.py", "noxfile.py", "src/pytestshellutils/"]
    _lint(session, ".pylintrc", flags, paths)


@nox.session(python="3", name="lint-tests")
def lint_tests(session):
    """
    Run PyLint against Salt and it's test suite. Set PYLINT_REPORT to a path to capture output.
    """
    flags = ["--disable=I"]
    if session.posargs:
        paths = session.posargs
    else:
        paths = ["tests/"]
    _lint(session, ".pylintrc", flags, paths)


@nox.session(python="3")
def docs(session):
    """
    Build Docs.
    """
    session.install("--progress-bar=off", "-e", ".[docs]", silent=PIP_INSTALL_SILENT)
    os.chdir("docs/")
    session.run("make", "clean", external=True)
    # session.run("make", "linkcheck", "SPHINXOPTS=-W", external=True)
    session.run("make", "coverage", "SPHINXOPTS=-W", external=True)
    docs_coverage_file = os.path.join("_build", "html", "python.txt")
    if os.path.exists(docs_coverage_file):
        with open(docs_coverage_file, encoding="utf-8") as rfh:
            contents = rfh.readlines()[2:]
            if contents:
                session.error("\n" + "".join(contents))
    session.run("make", "html", "SPHINXOPTS=-W", external=True)
    os.chdir("..")


@nox.session(name="docs-dev", python="3")
def docs_dev(session):
    """
    Build Docs.
    """
    session.install("--progress-bar=off", "-e", ".[docs]", silent=PIP_INSTALL_SILENT)
    os.chdir("docs/")
    session.run("make", "html", "SPHINXOPTS=-W", external=True, env={"LOCAL_DEV_BUILD": "1"})
    os.chdir("..")


@nox.session(name="docs-crosslink-info", python="3")
def docs_crosslink_info(session):
    """
    Report intersphinx cross links information.
    """
    session.install("--progress-bar=off", "-e", ".[docs]", silent=PIP_INSTALL_SILENT)
    os.chdir("docs/")
    intersphinx_mapping = json.loads(
        session.run(
            "python",
            "-c",
            "import json; import conf; print(json.dumps(conf.intersphinx_mapping))",
            silent=True,
            log=False,
        )
    )
    try:
        mapping_entry = intersphinx_mapping[session.posargs[0]]
    except IndexError:
        session.error(
            "You need to pass at least one argument whose value must be one of: {}".format(
                ", ".join(list(intersphinx_mapping))
            )
        )
    except KeyError:
        session.error(
            "Only acceptable values for first argument are: {}".format(
                ", ".join(list(intersphinx_mapping))
            )
        )
    session.run(
        "python", "-m", "sphinx.ext.intersphinx", mapping_entry[0].rstrip("/") + "/objects.inv"
    )
    os.chdir("..")


@nox.session(name="gen-api-docs", python="3")
def gen_api_docs(session):
    """
    Generate API Docs.
    """
    session.install("--progress-bar=off", "-e", ".[docs]", silent=PIP_INSTALL_SILENT)
    shutil.rmtree("docs/ref", ignore_errors=True)
    session.run("sphinx-apidoc", "--module-first", "-o", "docs/ref/", "src/pytestshellutils/")


@nox.session(name="twine-check", python="3")
def twine_check(session):
    """
    Run ``twine-check`` against the source distribution package.
    """
    build(session)
    session.run("twine", "check", "dist/*")


@nox.session(name="changelog", python="3")
@nox.parametrize("draft", [False, True])
def changelog(session, draft):
    """
    Generate changelog.
    """
    session.install("--progress-bar=off", "-e", ".[changelog]", silent=PIP_INSTALL_SILENT)

    version = session.run(
        "python",
        "setup.py",
        "--version",
        silent=True,
        log=False,
        stderr=None,
    ).strip()

    town_cmd = ["towncrier", "build", "--version={}".format(version)]
    if draft:
        town_cmd.append("--draft")
    session.run(*town_cmd)


@nox.session(name="release")
def release(session):
    """
    Create a release tag.
    """
    if not session.posargs:
        session.error(
            "Forgot to pass the version to release? For example `nox -e release -- 1.1.0`"
        )
    if len(session.posargs) > 1:
        session.error(
            "Only one argument is supported by the `release` nox session. "
            "For example `nox -e release -- 1.1.0`"
        )
    version = session.posargs[0]
    try:
        session.log("Generating temporary %s tag", version)
        session.run("git", "tag", "-as", version, "-m", "Release {}".format(version), external=True)
        changelog(session, draft=False)
    except CommandFailed:
        session.error("Failed to generate the temporary tag")
    # session.notify("changelog(draft=False)")
    try:
        session.log("Generating the release changelog")
        session.run(
            "git",
            "commit",
            "-a",
            "-m",
            "Generate Changelog for version {}".format(version),
            external=True,
        )
    except CommandFailed:
        session.error("Failed to generate the release changelog")
    try:
        session.log("Overwriting temporary %s tag", version)
        session.run(
            "git", "tag", "-fas", version, "-m", "Release {}".format(version), external=True
        )
    except CommandFailed:
        session.error("Failed to overwrite the temporary tag")
    session.warn("Don't forget to push the newly created tag")


class Recompress:
    """
    Helper class to re-compress a ``.tag.gz`` file to make it reproducible.
    """

    def __init__(self, mtime):
        self.mtime = int(mtime)

    def tar_reset(self, tarinfo):
        """
        Reset user, group, mtime, and mode to create reproducible tar.
        """
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = "root"
        tarinfo.mtime = self.mtime
        if tarinfo.type == tarfile.DIRTYPE:
            tarinfo.mode = 0o755
        else:
            tarinfo.mode = 0o644
        if tarinfo.pax_headers:
            raise ValueError(tarinfo.name, tarinfo.pax_headers)
        return tarinfo

    def recompress(self, targz):
        """
        Re-compress the passed path.
        """
        tempd = pathlib.Path(tempfile.mkdtemp()).resolve()
        d_src = tempd.joinpath("src")
        d_src.mkdir()
        d_tar = tempd.joinpath(targz.stem)
        d_targz = tempd.joinpath(targz.name)
        with tarfile.open(d_tar, "w|") as wfile:
            with tarfile.open(targz, "r:gz") as rfile:
                rfile.extractall(d_src)  # nosec
                extracted_dir = next(pathlib.Path(d_src).iterdir())
                for name in sorted(extracted_dir.rglob("*")):
                    wfile.add(
                        str(name),
                        filter=self.tar_reset,
                        recursive=False,
                        arcname=str(name.relative_to(d_src)),
                    )

        with open(d_tar, "rb") as rfh:
            with gzip.GzipFile(
                fileobj=open(d_targz, "wb"), mode="wb", filename="", mtime=self.mtime
            ) as gz:
                while True:
                    chunk = rfh.read(1024)
                    if not chunk:
                        break
                    gz.write(chunk)
        targz.unlink()
        shutil.move(str(d_targz), str(targz))


@nox.session(python="3")
def build(session):
    """
    Build source and binary distributions based off the current commit author date UNIX timestamp.

    The reason being, reproducible packages.

    .. code-block: shell

        git show -s --format=%at HEAD
    """
    shutil.rmtree("dist/", ignore_errors=True)
    session.install("--progress-bar=off", "-r", "requirements/build.txt", silent=PIP_INSTALL_SILENT)

    timestamp = session.run(
        "git",
        "show",
        "-s",
        "--format=%at",
        "HEAD",
        silent=True,
        log=False,
        stderr=None,
    ).strip()
    env = {"SOURCE_DATE_EPOCH": str(timestamp)}
    session.run(
        "python",
        "-m",
        "build",
        "--sdist",
        "--wheel",
        str(REPO_ROOT),
        env=env,
    )
    # Recreate sdist to be reproducible
    recompress = Recompress(timestamp)
    for targz in REPO_ROOT.joinpath("dist").glob("*.tar.gz"):
        session.log("Re-compressing %s...", targz.relative_to(REPO_ROOT))
        recompress.recompress(targz)

    sha256sum = shutil.which("sha256sum")
    if sha256sum:
        packages = [str(pkg.relative_to(REPO_ROOT)) for pkg in REPO_ROOT.joinpath("dist").iterdir()]
        session.run("sha256sum", *packages, external=True)
    session.run("python", "-m", "twine", "check", "dist/*")



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Test ``pytestshellutils.utils.processes.ProcessResult``.
"""
import pytest

from pytestshellutils.utils.processes import ProcessResult


MARY_HAD_A_LITTLE_LAMB = """\
Mary had a little lamb,
Its fleece was white as snow;
And everywhere that Mary went
The lamb was sure to go.
"""


@pytest.fixture
def process_result() -> ProcessResult:
    return ProcessResult(returncode=0, stdout=MARY_HAD_A_LITTLE_LAMB, stderr=None)


def test_instance_types() -> None:
    ret = ProcessResult(returncode=0, stdout="STDOUT", stderr="STDERR")
    assert isinstance(ret.stdout, str)
    assert isinstance(ret.stderr, str)
    ret = ProcessResult(returncode=0, stdout=None, stderr=None)
    assert ret.stdout is None
    assert ret.stderr is None
    ret = ProcessResult(returncode=0, stdout=1, stderr=2, data=None)
    assert ret.stdout == 1
    assert ret.stderr == 2


def test_matcher_attribute(process_result: ProcessResult) -> None:
    process_result.stdout.matcher.fnmatch_lines_random(
        [
            "*had a little*",
            "Its fleece was white*",
            "*Mary went",
            "The lamb was sure to go.",
        ]
    )



# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
#
import inspect
import pathlib
import sys
import warnings
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

import packaging.version

import pytestshellutils

if TYPE_CHECKING:
    from packaging.version import Version


def resolved_pathlib_path(path: Union[str, pathlib.Path]) -> pathlib.Path:
    """
    Return a resolved ``pathlib.Path``.
    """
    if isinstance(path, str):
        path = pathlib.Path(path)
    return path.resolve()


def format_callback_to_string(
    callback: Union[str, Callable[..., Any]],
    args: Optional[Tuple[Any, ...]] = None,
    kwargs: Optional[Dict[str, Any]] = None,
) -> str:
    """
    Convert a callback, its arguments and keyword arguments to a string suitable for logging purposes.

    Arguments:
        callback:
            The callback function

    Keyword Arguments:
        args:
            The callback arguments
        kwargs:
            The callback keyword arguments

    Returns:
        str: The formatted callback string
    """
    callback_str: str
    if not isinstance(callback, str):
        try:
            callback_str = f"{callback.__qualname__}("
        except AttributeError:  # pragma: no cover
            callback_str = f"{callback.__name__}("
    else:
        callback_str = f"{callback}("
    if args:
        callback_str += ", ".join([repr(arg) for arg in args])
    if kwargs:
        if args:
            callback_str += ", "
        callback_str += ", ".join([f"{k}={v!r}" for (k, v) in kwargs.items()])
    callback_str += ")"
    return callback_str


def warn_until(  # pragma: no cover
    version: str,
    message: str,
    category: Type[Warning] = DeprecationWarning,
    stacklevel: Optional[int] = None,
    _dont_call_warnings: bool = False,
    _pkg_version_: Optional[str] = None,
) -> None:
    """
    Show a deprecation warning.

    Helper function to raise a warning, by default, a ``DeprecationWarning``,
    until the provided ``version``, after which, a ``RuntimeError`` will
    be raised to remind the developers to remove the warning because the
    target version has been reached.

    Arguments:
        version:
            The version string after which the warning becomes a ``RuntimeError``.
            For example ``2.1``.
        message:
            The warning message to be displayed.

    Keyword Arguments:
        category:
            The warning class to be thrown, by default ``DeprecationWarning``
        stacklevel:
            There should be no need to set the value of ``stacklevel``.
        _dont_call_warnings:
            This parameter is used just to get the functionality until the actual
            error is to be issued. When we're only after the version checks to
            raise a ``RuntimeError``.

    Returns:
        Nothing.
    """
    _version = packaging.version.parse(version)
    if _pkg_version_ is None:
        _pkg_version_ = pytestshellutils.__version__  # type: ignore[attr-defined]
    _pkg_version = packaging.version.parse(_pkg_version_)

    if stacklevel is None:
        # Attribute the warning to the calling function, not to warn_until()
        stacklevel = 3

    if _pkg_version >= _version:
        caller = inspect.getframeinfo(sys._getframe(stacklevel - 1))
        raise RuntimeError(
            "The warning triggered on filename '{filename}', line number "
            "{lineno}, is supposed to be shown until version "
            "{until_version} is released. Current version is now "
            "{version}. Please remove the warning.".format(
                filename=caller.filename,
                lineno=caller.lineno,
                until_version=_pkg_version_,
                version=version,
            ),
        )

    if _dont_call_warnings is False:
        warnings.warn(
            message.format(version=version),
            category,
            stacklevel=stacklevel,
        )
